from string import ascii_letters
from unittest.mock import MagicMock

import mozanalysis.frequentist_stats.linear_models as mafslm
import numpy as np
import pandas as pd
import statsmodels.formula.api as smf

from .test_linear_model_functions import test_model, test_model_covariate


def test_from_formula():
    # no covariate

    actual = mafslm.MozOLS.from_formula(test_model.formula, test_model.model_df)

    # test that constructed design matrices are identical
    assert np.isclose(actual.exog, test_model.results.model.exog).all()
    assert np.isclose(actual.endog, test_model.results.model.endog).all()

    assert actual.formula == test_model.formula

    pd.testing.assert_frame_equal(actual.data.frame, test_model.model_df)

    assert actual.k_constant == test_model.results.model.k_constant

    # with covariate

    actual = mafslm.MozOLS.from_formula(
        test_model_covariate.formula, test_model_covariate.model_df
    )

    assert np.isclose(actual.exog, test_model_covariate.results.model.exog).all()
    assert np.isclose(actual.endog, test_model_covariate.results.model.endog).all()

    assert actual.formula == test_model_covariate.formula

    pd.testing.assert_frame_equal(actual.data.frame, test_model_covariate.model_df)

    assert actual.k_constant == test_model_covariate.results.model.k_constant


def _compare_models(actual, expected):
    assert np.isclose(actual.params, expected.params).all()
    assert np.isclose(actual.cov_params(), expected.cov_params()).all()
    assert np.isclose(
        actual.normalized_cov_params, expected.normalized_cov_params
    ).all()
    assert np.isclose(actual.bse, expected.bse).all()

    assert np.isclose(actual.llf, expected.llf)


def test_fit_basic(monkeypatch):
    # since sm.OLS inherts a fit method from statsmodels.base.LikelihoodModel
    # these checks ensure that MozOLS calls our fit method, not the original
    pinv_mock = MagicMock()
    monkeypatch.setattr("numpy.linalg.pinv", pinv_mock)
    # no covariate
    model = mafslm.MozOLS.from_formula(test_model.formula, test_model.model_df)
    actual = model.fit()
    pinv_mock.assert_not_called()

    _compare_models(actual, test_model.results)

    # with covariate
    model = mafslm.MozOLS.from_formula(
        test_model_covariate.formula, test_model_covariate.model_df
    )
    actual = model.fit()
    pinv_mock.assert_not_called()

    _compare_models(actual, test_model_covariate.results)


def test_fit_random():
    np.random.seed = 42
    for n_col in range(2, 10):
        for n_row in range(100, 10_000, 1000):
            model_df = pd.DataFrame(
                np.random.random(size=(n_row, n_col)),
                columns=list(ascii_letters[:n_col]),
            )
            formula = "a ~ "
            for letter in ascii_letters[1:n_col]:
                formula += f"{letter} + "
            formula = formula[:-3]

            expected = smf.ols(formula, model_df).fit()
            actual = mafslm.MozOLS.from_formula(formula, model_df).fit()

            _compare_models(actual, expected)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""module for sample size calculations"""

from collections import UserDict
from datetime import datetime
from math import pi

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas.io.formats.style import Styler
from scipy.stats import norm
from statsmodels.stats.power import tt_ind_solve_power, zt_ind_solve_power
from statsmodels.stats.proportion import samplesize_proportions_2indep_onetail

from mozanalysis.bq import BigQueryContext
from mozanalysis.experiment import TimeSeriesResult
from mozanalysis.metrics import Metric
from mozanalysis.segments import Segment
from mozanalysis.sizing import HistoricalTarget
from mozanalysis.utils import get_time_intervals


class ResultsHolder(UserDict):
    """
    Object to hold results from different methods.  It extends
    the dictionary objects so that users can interact with it
    like a dictionary with the same keys/values as before, making
    it backward compatible.
    """

    def __init__(
        self, *args, metrics: dict | None = None, params: dict | None = None, **kwargs
    ):
        """
        Args:
            metrics (list, optional): List of metrics used to generate the results.
                                     Defaults to None.
            params (dict, optional): Parameters used to generated the results.
                                    Defaults to None.
        """
        super().__init__(*args, **kwargs)
        self._metrics = metrics
        self._params = params
        # create this attribute to hold only the results data
        # this dict is what was returned from the sample size
        # methods historically
        self.data = {
            k: v for k, v in self.data.items() if k not in ["metrics", "params"]
        }

    @property
    def metrics(self):
        """List of metrics used to generate the results. Defaults to None"""
        return self._metrics

    @property
    def params(self):
        """Parameters used to generated the results.  Defaults to None"""
        return self._params

    @staticmethod
    def make_friendly_name(ugly_name: str) -> str:
        """Turns a name into a friendly name
        by replacing underscores with spaces and
        capitlizing words other than ones like "per", "of", etc
        Args:
            ugly_name (str): name to make pretty

        Returns:
            pretty_name (str): reformatted name
        """
        keep_all_lowercase = ["per", "of"]
        split_name = ugly_name.split("_")
        split_name = [
            el[0].upper() + el[1:] if el not in keep_all_lowercase else el
            for el in split_name
        ]
        pretty_name = " ".join(split_name)
        return pretty_name


class SampleSizeResultsHolder(ResultsHolder):
    """
    Object to hold results from different methods.  It extends
    the dictionary objects so that users can interact with it
    with a dictionary with the same keys/values as before, making
    it backward compatible.  The dictionary functionality is extended to include
    additional attributes to hold metadata, a method for plotting results and
    a method for returning results as a dataframe
    """

    @property
    def dataframe(self) -> pd.DataFrame:
        """dataframe property

        returns data as a dataframe rather than a dict
        """
        return pd.DataFrame(self.data).transpose()

    def plot_results(self, result_name: str = "sample_size_per_branch"):
        """plots the outputs of the sampling methods

        Args:
            result_name (str): sample size method output to plot.
            Defaults to sample_size_per_branch
        """
        nice_metric_map = {
            el.name: (
                el.friendly_name
                if hasattr(el, "friendly_name")
                else self.make_friendly_name(el.name)
            )
            for el in self._metrics
        }
        nice_result = self.make_friendly_name(result_name)
        df = self.dataframe.rename(index=nice_metric_map)
        df[result_name].plot(kind="bar")
        plt.ylabel(nice_result)
        plt.xlabel("Metric")
        plt.show(block=False)


class EmpiricalEffectSizeResultsHolder(ResultsHolder):
    "ResultsHolder for empirical_effect_size_sample_size_calc"

    def style_empirical_sizing_result(
        self, empirical_sizing_df: pd.DataFrame
    ) -> Styler:
        """Pretty-print the DF returned by `empirical_sizing()`.

        Returns a pandas Styler object.
        """
        return (
            empirical_sizing_df[
                [
                    "relative_effect_size",
                    "population_percent_per_branch",
                    "sample_size_per_branch",
                    "effect_size_value",
                    "mean_value",
                    "std_dev_value",
                ]
            ]
            .rename(
                columns={
                    "relative_effect_size": "rel_effect_size",
                    "population_percent_per_branch": "branch_pop_pct",
                    "sample_size_per_branch": "branch_sample_size",
                }
            )
            .style.format(thousands=",", precision=4)
            .format(subset="branch_sample_size", thousands=",", precision=0)
            .format("{:.2f}%", subset="branch_pop_pct")
            .format("{:.2%}", subset="rel_effect_size")
        )

    @property
    def dataframe(self) -> pd.DataFrame:
        """dataframe property

        returns results consolidated into a dataframe"""

        formatted_results = {}
        for m, r in self.data.items():
            metric_result = {}
            for k, v in r.items():
                if isinstance(v, dict):
                    for vk, vv in v.items():
                        if vk == "period_start_day":
                            vk = "period"
                        metric_result[f"{k}_{vk}"] = vv
                else:
                    metric_result[k] = v
            formatted_results[m] = metric_result

        df = pd.DataFrame.from_dict(formatted_results, orient="index")

        return df[
            [
                "relative_effect_size",
                "effect_size_value",
                "mean_value",
                "std_dev_value",
                "sample_size_per_branch",
                "population_percent_per_branch",
                "effect_size_period",
                "mean_period",
                "std_dev_period",
            ]
        ]

    def get_styled_dataframe(self) -> Styler:
        """returns styled dataframe for results from
        empirical_effect_size_sample_size_calc

        Arguments:
            style (bool): If true, return a cleaned up and formatted pandas Styler.
            Otherwise return a dataframe
        Returns:
            Styler: styled dataframe for visualization
        """
        df = self.dataframe
        return self.style_empirical_sizing_result(df)


class SampleSizeCurveResultHolder(ResultsHolder):
    def __init__(self, *args, **kwargs):
        # uses the __init__ method for ResultsHolder
        super().__init__(*args, **kwargs)
        # need to modify results so it matches the old format
        sim_var = self._params["sim_var"]
        self.raw_df = pd.concat(
            [el.set_index(sim_var, append=True) for el in self.data.values()]
        )
        metrics = [el.name for el in self._metrics]
        results_dict = {}
        for el in metrics:
            results_dict[el] = self.raw_df.loc[el, :].reset_index()
        self.data = results_dict

    def set_raw_data_stats(self, input_data: pd.DataFrame) -> pd.DataFrame:
        outlier_percentile = self._params["outlier_percentile"] / 100
        overall_stats = (
            input_data[[m.name for m in self._metrics]]
            .agg(
                [
                    "mean",
                    "std",
                    lambda d: d[d <= d.quantile(outlier_percentile)].mean(),
                    lambda d: d[d <= d.quantile(outlier_percentile)].std(),
                ]
            )
            .transpose()
        )
        overall_stats.columns = ["mean", "std", "mean_trimmed", "std_trimmed"]
        overall_stats["trim_change_mean"] = (
            overall_stats["mean_trimmed"] - overall_stats["mean"]
        ).abs() / overall_stats["mean"]
        overall_stats["trim_change_std"] = (
            overall_stats["std_trimmed"] - overall_stats["std"]
        ).abs() / overall_stats["std"]
        self._raw_data_stats = overall_stats

    @property
    def dataframe(self) -> pd.DataFrame:
        """dataframe property"""
        return self.raw_df

    def get_styled_dataframe(
        self,
        input_data: pd.DataFrame = None,
        show_population_pct: bool = True,
        simulated_values: list[float] | None = None,
        append_stats: bool = False,
        highlight_lessthan: list[float] | None = None,
        trim_highlight_threshold: float = 0.15,
    ) -> Styler:
        """
        Returns styled dataframe useful for visualization

        Args:
            input_data (pd.DataFrame, optional): Metric data used for summary stats.
                Defaults to None.
            show_population_pct (bool, optional): Controls whether output is a percent
                of population or a count. Defaults to True.
            simulated_values (List[float], optional): List of values that were varied
                to create curves. Defaults to None.
            append_stats (bool, optional): Controls whether or not to append summary
                stats to output dataframe. Defaults to False.
            highlight_lessthan (List[float], optional): list of sample size thresholds
                to highlight in the results. For each threshold, sample sizes lower than
                it (but higher than any other thresholds) are highlighted in a
                predefined colour. When `show_population_pct` is `True`, thresholds
                should be expressed as a percentage between 0 and 100, not a decimal
                between 0 and 1 (for example, to set a threshold for 5%, supply `[5]`).
                At most 3 different thresholds are supported: only the 3 lowest
                thresholds supplied will be used, and any others are silently ignored.
                Defaults to None.
            trim_highlight_threshold (float, optional): if summary stats are shown,
                cases for which the trimmed mean differs from the raw mean by more
                than this threshold are highlighted. These metrics are strongly affected
                by outliers. The threshold should be a relative difference value between
                0 and 1.. Defaults to 0.15.

        Returns:
            Styler: styled output
        """
        # choose which column to output
        if show_population_pct:
            subset_col = "population_percent_per_branch"
        else:
            subset_col = "sample_size_per_branch"

        # reformate raw_df to make different simulatd values into columns
        if simulated_values is None:
            simulated_values = self._params["simulated_values"]
            pretty_df = self.raw_df[subset_col].unstack()
        else:
            pretty_df = self.raw_df[subset_col].unstack()[simulated_values]

        # get stats if input_data is non-null and append_stats is True
        if append_stats:
            if input_data is None:
                raise ValueError("append_stats is true but no raw data was provided")
            self.set_raw_data_stats(input_data)
            pretty_df = pd.concat([pretty_df, self._raw_data_stats], axis="columns")

        disp = (
            pretty_df.style.format(
                "{:.2f}%" if show_population_pct else "{:,.0f}", subset=simulated_values
            )
            # Round displayed effect size values to smallest precision for readability
            .format_index(
                axis="columns",
                precision=int(-np.floor(np.log10(simulated_values)).min()),
            )
        )
        if append_stats:
            large_change_format_str = "color:maroon; font-weight:bold"
            disp = (
                disp.set_properties(
                    subset=self._raw_data_stats.columns,
                    **{"background-color": "tan", "color": "black"},
                )
                .format("{:.2%}", subset=["trim_change_mean", "trim_change_std"])
                # highlight large changes in mean because of trimming
                .applymap(
                    lambda x: (
                        large_change_format_str if x > trim_highlight_threshold else ""
                    ),
                    subset=["trim_change_mean"],
                )
            )

        # Colours chosen to work reasonably well in either light or dark mode
        # Ordered from highest to lowest cutoff
        highlight_colours = ["lightgreen", "skyblue", "gold"]
        if highlight_lessthan:
            # Only 3 cutoffs suported. Any others are silently dropped
            highlight_lessthan = sorted(highlight_lessthan)[:3]
            # Apply from highest to lowest cutoff so that lower cutoffs overwrite higher
            for lim, colour in list(
                zip(highlight_lessthan, highlight_colours, strict=False)
            )[::-1]:
                disp = disp.highlight_between(
                    subset=simulated_values,
                    right=lim,
                    props=f"background-color:{colour};color:black",
                )

        return disp


def sample_size_curves(
    df: pd.DataFrame,
    metrics_list: list,
    solver,
    effect_size: float | np.ndarray | pd.Series | list[float] = 0.01,
    power: float | np.ndarray | pd.Series | list[float] = 0.80,
    alpha: float | np.ndarray | pd.Series | list[float] = 0.05,
    **solver_kwargs,
) -> SampleSizeCurveResultHolder:
    """
    Loop over a list of different parameters to produce sample size estimates given
    those parameters. A single parameter in [effect_size, power, alpha] should
    be passed a list; the sample size curve will be calculated with this as
    the variable.

    Args:
        df: A pandas DataFrame of queried historical data.
        metrics_list (list of mozanalysis.metrics.Metric): List of metrics
            used to construct the results df from HistoricalTarget. The names
            of these metrics are used to return results for sample size
            calculation for each.
        solver (any function that returns sample size as function of
            effect_size, power, alpha): The solver being used to calculate sample
            size.
        effect_size (float or ArrayLike, default .01): For test of differences in
            proportions, the absolute difference; for tests of differences in mean,
            the percent change.
        alpha (float or ArrayLike, default .05): Significance level for the experiment.
        power (float or ArrayLike, default .90): Probability of detecting an effect,
            when a significant effect exists.
        **solver_kwargs (dict): Arguments necessary for the provided solver.

    Returns:
        SampleSizeCurveResultHolder: The data attribute contains a dictionary
        of pd.DataFrame objects. An item in the dictionary is
        created for each metric in metric_list, containing a DataFrame of sample
        size per branch, number of clients that satisfied targeting, and population
        proportion per branch at each value of the iterable parameter.
        Additional methods for ease of use are documented in the class.
    """
    params = {"effect_size": effect_size, "power": power, "alpha": alpha}
    sim_var = [k for k, v in params.items() if type(v) in [list, np.ndarray, pd.Series]]

    if len(sim_var) != 1:
        raise ValueError(
            "Exactly one of effect_size, power, and alpha must be ArrayLike"
        )

    sim_var = sim_var[0]
    test_vals = params[sim_var]
    del params[sim_var]
    results = {}
    for v in test_vals:
        sample_sizes = solver(
            df, metrics_list, **{sim_var: v}, **params, **solver_kwargs
        ).dataframe
        sample_sizes[sim_var] = v

        results[v] = sample_sizes

    # add sim_var to metadata
    params["sim_var"] = sim_var
    params["simulated_values"] = test_vals
    return SampleSizeCurveResultHolder(
        results, metrics=metrics_list, params={**params, **solver_kwargs}
    )


def difference_of_proportions_sample_size_calc(
    df: pd.DataFrame,
    metrics_list: list[Metric],
    effect_size: float = 0.01,
    alpha: float = 0.05,
    power: float = 0.90,
    outlier_percentile: float = 99.5,
) -> SampleSizeResultsHolder:
    """
    Perform sample size calculation for an experiment to test for a
    difference in proportions.

    Args:
        df: A pandas DataFrame of queried historical data.
        metrics_list (list of mozanalysis.metrics.Metric): List of metrics
            used to construct the results df from HistoricalTarget. The names
            of these metrics are used to return results for sample size
            calculation for each
        effect_size (float, default .01): Difference in proportion for the
            minimum detectable effect --
            effect_size = p(event under alt) - p(event under null)
        alpha (float, default .05): Significance level for the experiment.
        power (float, default .90): Probability of detecting an effect,
            when a significant effect exists.
        outlier_percentile(float, default .995): Percentile at which to trim
            each columns.

    Returns:
        SampleSizeResultsHolder: The data attribute contains a dictionary.
        Keys in the dictionary are the metrics column names from
        the DataFrame; values are the required sample size per branch to achieve
        the desired power for that metric. Additional methods for ease of use
        are documented in the class.
    """

    def _get_sample_size_col(col):
        p = np.percentile(df[col], q=[outlier_percentile])[0]
        mean = df.loc[df[col] <= p, col].mean()
        p2 = mean + effect_size

        return samplesize_proportions_2indep_onetail(
            diff=effect_size, prop2=p2, power=power, ratio=1, alpha=alpha, value=0
        )

    metric_names = [m.name for m in metrics_list]
    results = {}
    for col in metric_names:
        sample_size = _get_sample_size_col(col)
        pop_percent = 100.0 * (sample_size / len(df))
        results[col] = {
            "sample_size_per_branch": sample_size,
            "population_percent_per_branch": pop_percent,
            "number_of_clients_targeted": len(df),
        }
    params = {
        "effect_size": effect_size,
        "alpha": alpha,
        "power": power,
        "outlier_percentile": outlier_percentile,
    }

    return SampleSizeResultsHolder(results, metrics=metrics_list, params=params)


def z_or_t_ind_sample_size_calc(
    df: pd.DataFrame,
    metrics_list: list[Metric],
    test: str = "z",
    effect_size: float = 0.01,
    alpha: float = 0.05,
    power: float = 0.90,
    outlier_percentile: float = 99.5,
) -> SampleSizeResultsHolder:
    """
    Perform sample size calculation for an experiment based on independent
    samples t or z tests.

    Args:
        df: A pandas DataFrame of queried historical data.
        metrics_list (list of mozanalysis.metrics.Metric): List of metrics
            used to construct the results df from HistoricalTarget. The names
            of these metrics are used to return results for sample size
            calculation for each
        test (str, default `z`): `z` or `t` to indicate which solver to use
        effect_size (float, default .01): Percent change in metrics
            expected as a result of the experiment treatment
        alpha (float, default .05): Significance level for the experiment.
        power (float, default .90): Probability of detecting an effect,
            when a significant effect exists.
        outlier_percentile(float, default .995): Percentile at which to trim
            each columns.

    Returns:
        SampleSizeResultsHolder: The data attribute contains a dictionary.
        Keys in the dictionary are the metrics column names from
        the DataFrame; values are the required sample size per branch to achieve
        the desired power for that metric. Additional methods for ease of use
        are documented in the class.
    """
    tests = {
        "normal": zt_ind_solve_power,
        "z": zt_ind_solve_power,
        "t": tt_ind_solve_power,
    }
    solver = tests[test]

    def _get_sample_size_col(col):
        p = np.percentile(df[col], q=[outlier_percentile])[0]
        sd = df.loc[df[col] <= p, col].std()
        mean = df.loc[df[col] <= p, col].mean()
        es = (effect_size * mean) / sd

        return solver(effect_size=es, alpha=alpha, power=power, nobs1=None)

    metric_names = [m.name for m in metrics_list]
    results = {}
    for col in metric_names:
        sample_size = _get_sample_size_col(col)
        pop_percent = 100.0 * (sample_size / len(df))
        results[col] = {
            "sample_size_per_branch": sample_size,
            "population_percent_per_branch": pop_percent,
            "number_of_clients_targeted": len(df),
        }
    params = {
        "effect_size": effect_size,
        "alpha": alpha,
        "power": power,
        "outlier_percentile": outlier_percentile,
        "solver": solver,
        "test": test,
    }

    return SampleSizeResultsHolder(results, metrics=metrics_list, params=params)


def empirical_effect_size_sample_size_calc(
    res: TimeSeriesResult,
    bq_context: BigQueryContext,
    metric_list: list,
    quantile: float = 0.90,
    power: float = 0.80,
    alpha: float = 0.05,
    parent_distribution: str = "normal",
    plot_effect_sizes: bool = False,
) -> EmpiricalEffectSizeResultsHolder:
    """
    Perform sample size calculation with empirical effect size and
    asymptotic approximation of Wilcoxen-Mann-Whitney U Test. Empirical effect size
    is estimated using a quantile of week-to-week changes over
    the course of the study, and the variance in the test statistic is
    estimated as a quantile of weekly variance in metrics. Sample
    size calculation is based on the asymptotic relative efficiency (ARE) of
    the U test to the T test (see Stapleton 2008, pg 266, or
    https://www.psychologie.hhu.de/fileadmin/redaktion/Fakultaeten/
    Mathematisch-Naturwissenschaftliche_Fakultaet/Psychologie/AAP/gpower/GPowerManual.pdf)

    Args:
        res: A TimeSeriesResult, generated by
            mozanalysis.sizing.HistoricalTarget.get_time_series_data.
        bq_context: A mozanalysis.bq.BigQueryContext object that handles downloading
            time series data from BigQuery.
        metrics_list (list of mozanalysis.metrics.Metric): List of metrics
            used to construct the results df from HistoricalTarget. The names
            of these metrics are used to return results for sample size
            calculation for each.
        quantile (float, default .90): Quantile used to calculate the effect size
            as the quantile of week-to-week metric changes and the variance of
            the mean.
        alpha (float, default .05): Significance level for the experiment.
        power (float, default .90): Probability of detecting an effect,
            when a significant effect exists.
        parent_distribution (str, default "normal"): Distribution of the parent data;
            must be normal, uniform, logistic, or laplace.
        plot_effect_sizes (bool, default False): Whether or not to plot the
            distribution of effect sizes observed in historical data.

    Returns:
        EmpiricalEffectSizeResultsHolder: The data attribute contains a dictionary.
        Keys in the dictionary are the metrics column names from
        the DataFrame; values are dictionaries containing the required sample size
        per branch to achieve the desired power for that metric, along with
        additional information. Additional methods for ease of use
        are documented in the class.
    """

    def _mann_whitney_solve_sample_size_approximation(
        effect_size, std, alpha=0.05, power=0.8, parent_distribution="normal"
    ):
        """
        Wilcoxen-Mann-Whitney rank sum test sample size calculation,
        based on asymptotic efficiency relative to the t-test.
        """
        rel_effect_size = effect_size / std
        are = {
            "uniform": 1.0,
            "normal": pi / 3.0,
            "logistic": 9.0 / (pi**2),
            "laplace": 2.0 / 3.0,
        }

        if parent_distribution not in are:
            raise ValueError(f"Parent distribution must be in {are.keys()}")

        t_sample_size = tt_ind_solve_power(
            effect_size=rel_effect_size, power=power, alpha=alpha
        )

        return t_sample_size * are[parent_distribution]

    res_mean, pop_size = res.get_aggregated_data(
        bq_context=bq_context, metric_list=metric_list, aggregate_function="AVG"
    )
    res_mean.sort_values(by="analysis_window_start", ascending=True, inplace=True)

    res_std, _ = res.get_aggregated_data(
        bq_context=bq_context, metric_list=metric_list, aggregate_function="STDDEV"
    )

    size_dict = {}

    for m in metric_list:
        res_mean["diff"] = res_mean[m.name].diff().abs()
        if plot_effect_sizes:
            print(f"{m.name}: plotting effect sizes observed in historical data")
            print("Summary statistics")
            print(res_mean["diff"].describe())
            print("Histogram of effect sizes")
            plt.hist(res_mean["diff"], bins=20)
            plt.show()
        m_quantile = res_mean["diff"].quantile(q=quantile, interpolation="nearest")
        m_std = res_std[m.name].quantile(q=quantile, interpolation="nearest")

        effect_size = {
            "value": m_quantile,
            "period_start_day": res_mean.loc[
                res_mean["diff"] == m_quantile, "analysis_window_start"
            ].values[0],
        }
        effect_size_base_period = effect_size["period_start_day"] - 7
        metric_value = {
            "value": res_mean.loc[
                res_mean["analysis_window_start"] == effect_size_base_period, m.name
            ].values[0],
            "period_start_day": effect_size_base_period,
        }
        std = {
            "value": m_std,
            "period_start_day": res_std.loc[
                res_std[m.name] == m_std, "analysis_window_start"
            ].values[0],
        }
        sample_size = _mann_whitney_solve_sample_size_approximation(
            effect_size=effect_size["value"],
            std=std["value"],
            power=power,
            alpha=alpha,
            parent_distribution=parent_distribution,
        )
        size_dict[m.name] = {
            "effect_size": effect_size,
            "mean": metric_value,
            "std_dev": std,
            "relative_effect_size": effect_size["value"] / metric_value["value"],
            "sample_size_per_branch": sample_size,
            "number_of_clients_targeted": pop_size,
            "population_percent_per_branch": 100.0 * (sample_size / pop_size),
        }

    params = {
        "quantile": quantile,
        "power": power,
        "alpha": alpha,
        "parent_distribution": parent_distribution,
    }

    return EmpiricalEffectSizeResultsHolder(
        size_dict, metrics=metric_list, params=params
    )


def poisson_diff_solve_sample_size(
    df: pd.DataFrame,
    metrics_list: list[Metric],
    effect_size: float = 0.01,
    alpha: float = 0.05,
    power: float = 0.90,
    outlier_percentile: float = 99.5,
) -> SampleSizeResultsHolder:
    """
    Sample size for test of difference of Poisson rates,
    based on Poisson rate's asymptotic normality.

    Args:
        df: A pandas DataFrame of queried historical data.
        metrics_list (list of mozanalysis.metrics.Metric): List of metrics
            used to construct the results df from HistoricalTarget. The names
            of these metrics are used to return results for sample size
            calculation for each
        test (str, default `z`): `z` or `t` to indicate which solver to use
        effect_size (float, default .01): Percent change in metrics
            expected as a result of the experiment treatment
        alpha (float, default .05): Significance level for the experiment.
        power (float, default .90): Probability of detecting an effect,
            when a significant effect exists.
        outlier_percentile(float, default .995): Percentile at which to trim
            each columns.

    Returns:
        SampleSizeResultsHolder: The data attribute contains a dictionary.
        Keys in the dictionary are the metrics column names from
        the DataFrame; values are the required sample size per branch to achieve
        the desired power for that metric.  Additional methods for ease of use
        are documented in the class
    """

    def _get_sample_size_col(col):
        p = np.percentile(df[col], q=[outlier_percentile])[0]
        sd = df.loc[df[col] <= p, col].std()
        mean = df.loc[df[col] <= p, col].mean()
        es = (effect_size * mean) / sd

        z_alpha = norm.ppf(1 - alpha / 2)
        z_power = norm.ppf(power)

        denom = (es / (z_alpha + z_power)) ** 2
        sample_size = (mean + es) / denom
        return sample_size

    metric_names = [m.name for m in metrics_list]
    results = {}
    for col in metric_names:
        sample_size = _get_sample_size_col(col)
        pop_percent = 100.0 * (sample_size / len(df))
        results[col] = {
            "sample_size_per_branch": sample_size,
            "population_percent_per_branch": pop_percent,
            "number_of_clients_targeted": len(df),
        }
    params = {
        "effect_size": effect_size,
        "alpha": alpha,
        "power": power,
        "outlier_percentile": outlier_percentile,
    }

    return SampleSizeResultsHolder(results, metrics=metrics_list, params=params)


def variable_enrollment_length_sample_size_calc(
    bq_context: BigQueryContext,
    start_date: str | datetime,
    max_enrollment_days: int,
    analysis_length: int,
    metric_list: list[Metric],
    target_list: list[Segment],
    variable_window_length: int = 7,
    experiment_name: str | None = "",
    app_id: str | None = "",
    to_pandas: bool = True,
    **sizing_kwargs,
) -> dict[str, dict[str, int] | pd.DataFrame]:
    """
    Sample size calculation over a variable enrollment window. This function
    will fetch a DataFrame with metrics defined in metric_list for a target
    population defined in the target_list over an enrollment window of length
    max_enrollment_days. Sample size calculation is performed
    using clients enrolled in the first variable_window_length dates in
    that max enrollment window; that window is incrementally widened by
    the variable window length and sample size calculation performed again,
    until the last enrollment date is reached.

    Args:
        bq_context: A mozanalysis.bq.BigQueryContext object that handles downloading
            data from BigQuery.
        start_date (str or datetime in %Y-%m-%d format): First date of enrollment for
            sizing job.
        max_enrollment_days (int): Maximum number of dates to consider for the
            enrollment period for the experiment in question.
        analysis_length (int): Number of days to record metrics for each client
            in the experiment in question.
        metric_list (list of mozanalysis.metrics.Metric): List of metrics
            used to construct the results df from HistoricalTarget. The names
            of these metrics are used to return results for sample size
            calculation for each.
        target_list (list of mozanalysis.segments.Segment): List of segments
            used to identify clients to include in the study.
        variable_window_length (int): Length of the intervals used to extend
            the enrollment period incrementally. Sample sizes are recalculated over
            each variable enrollment period.
        experiment_name (str): Optional name used to name the target and metric
            tables in BigQuery.
        app_id (str): Application that experiment will be run on.
        **sizing_kwargs: Arguments to pass to z_or_t_ind_sample_size_calc

    Returns:
        A dictionary. Keys in the dictionary are the metrics column names from
        the DataFrame; values are the required sample size per branch to achieve
        the desired power for that metric.
    """

    if variable_window_length > max_enrollment_days:
        raise ValueError(
            "Enrollment window length is larger than the max enrollment length."
        )

    ht = HistoricalTarget(
        start_date=start_date,
        analysis_length=analysis_length,
        num_dates_enrollment=max_enrollment_days,
        experiment_name=experiment_name,
        app_id=app_id,
    )

    df = ht.get_single_window_data(
        bq_context=bq_context, metric_list=metric_list, target_list=target_list
    )

    interval_end_dates = get_time_intervals(
        start_date,
        variable_window_length,
        max_enrollment_days,
    )

    def _for_interval_sample_size_calculation(i):
        df_interval = df.loc[df["enrollment_date"] < interval_end_dates[i]]
        res = z_or_t_ind_sample_size_calc(
            df=df_interval, metrics_list=metric_list, test="t", **sizing_kwargs
        )
        final_res = {}
        for key in res:
            final_res[key] = {
                "enrollment_end_date": interval_end_dates[i],
                **res[key],
            }

        return final_res

    results_dict = {}
    for m in metric_list:
        results_dict[m.name] = []

    for i in range(len(interval_end_dates)):
        res = _for_interval_sample_size_calculation(i)
        for m in metric_list:
            results_dict[m.name].append(res[m.name])

    for m in results_dict:
        results_dict[m] = pd.DataFrame(results_dict[m])

    return results_dict



from textwrap import dedent

import pytest
from helpers.cheap_lint import sql_lint  # local helper file
from helpers.config_loader_lists import (
    desktop_metrics,
    desktop_segments,
    fenix_metrics,
    firefox_ios_metrics,
    klar_android_metrics,
    klar_ios_metrics,
)
from metric_config_parser import AnalysisUnit
from mozanalysis.config import ApplicationNotFound, ConfigLoader
from mozanalysis.experiment import (
    AnalysisWindow,
    EnrollmentsQueryType,
    Experiment,
    IncompatibleAnalysisUnit,
    TimeLimits,
)
from mozanalysis.exposure import ExposureSignal
from mozanalysis.metrics import AnalysisBasis, DataSource, Metric
from mozanalysis.segments import Segment, SegmentDataSource


def test_time_limits_validates():
    # Mainly check that the validation is running at all
    # No need to specify the same checks twice(?)
    with pytest.raises(TypeError):
        TimeLimits()

    with pytest.raises(AssertionError):
        TimeLimits(
            first_enrollment_date="2019-01-05",
            last_enrollment_date="2019-01-05",
            analysis_windows=(AnalysisWindow(1, 1),),
            first_date_data_required="2019-01-01",  # Before enrollments
            last_date_data_required="2019-01-01",
        )


def test_time_limits_create1():
    # When we have complete data for 2019-01-14...
    # ...We have 14 dates of data for those who enrolled on the 1st
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=0,
        analysis_length_dates=14,
    )

    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-01"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 13
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_time_limits_create2():
    # We don't have 14 dates of data for an 8-day cohort:
    with pytest.raises(
        ValueError,
        match=(
            "You said you wanted 8 dates of enrollment, "
            + "and need data from the 13th day after enrollment. "
            + "For that, you need to wait until we have data for 2019-01-21."
        ),
    ):
        TimeLimits.for_single_analysis_window(
            first_enrollment_date="2019-01-01",
            last_date_full_data="2019-01-14",
            analysis_start_days=0,
            analysis_length_dates=14,
            num_dates_enrollment=8,
        )

    # We don't have 15 full dates of data for any users
    with pytest.raises(AssertionError):
        TimeLimits.for_single_analysis_window(
            first_enrollment_date="2019-01-01",
            last_date_full_data="2019-01-14",
            analysis_start_days=0,
            analysis_length_dates=15,
        )


def test_time_limits_create3():
    # For the 8-day cohort We have enough data for a 7 day window
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=0,
        analysis_length_dates=7,
        num_dates_enrollment=8,
    )
    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 6
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_time_limits_create4():
    # Or a 2 day window
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=0,
        analysis_length_dates=2,
        num_dates_enrollment=8,
    )
    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 1
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-09"


def test_time_limits_create5():
    # But not an 8 day window
    with pytest.raises(
        ValueError,
        match=(
            "You said you wanted 8 dates of enrollment, "
            + "and need data from the 7th day after enrollment. "
            + "For that, you need to wait until we have data for 2019-01-15."
        ),
    ):
        TimeLimits.for_single_analysis_window(
            first_enrollment_date="2019-01-01",
            last_date_full_data="2019-01-14",
            analysis_start_days=0,
            analysis_length_dates=8,
            num_dates_enrollment=8,
        )


def test_time_limits_create6():
    # Of course the flexi-experiment has data for a 1 day window
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=0,
        analysis_length_dates=1,
    )
    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-14"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 0
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_time_limits_create7():
    # If the analysis starts later, so does the data source
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=7,
        analysis_length_dates=1,
    )
    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-07"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 7
    assert tl.analysis_windows[0].end == 7
    assert tl.first_date_data_required == "2019-01-08"
    assert tl.last_date_data_required == "2019-01-14"


def test_ts_time_limits_create1():
    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        time_series_period="daily",
        num_dates_enrollment=8,
    )

    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 7
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 0
    assert tl.analysis_windows[6].start == 6
    assert tl.analysis_windows[6].end == 6
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_ts_time_limits_create2():
    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 6
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_ts_time_limits_create3():
    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-15",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 6
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-01-14"


def test_ts_time_limits_create4():
    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-02-15",
        time_series_period="28_day",
        num_dates_enrollment=8,
    )

    assert tl.first_enrollment_date == "2019-01-01"
    assert tl.last_enrollment_date == "2019-01-08"
    assert len(tl.analysis_windows) == 1
    assert tl.analysis_windows[0].start == 0
    assert tl.analysis_windows[0].end == 27
    assert tl.first_date_data_required == "2019-01-01"
    assert tl.last_date_data_required == "2019-02-04"


def test_ts_time_limits_create_not_enough_data():
    with pytest.raises(ValueError, match="Insufficient data"):
        TimeLimits.for_ts(
            first_enrollment_date="2019-01-01",
            last_date_full_data="2019-01-13",
            time_series_period="weekly",
            num_dates_enrollment=8,
        )


def test_time_limits_has_right_date_in_error_message():
    msg_re = r"until we have data for 2020-03-30."
    with pytest.raises(ValueError, match=msg_re):
        TimeLimits.for_single_analysis_window(
            first_enrollment_date="2020-03-03",
            last_date_full_data="2020-03-23",
            analysis_start_days=0,
            analysis_length_dates=21,
            num_dates_enrollment=8,
        )


def test_analysis_window_validates_start():
    AnalysisWindow(0, 1)
    AnalysisWindow(-2, -1)
    with pytest.raises(AssertionError):
        AnalysisWindow(-1, 1)


def test_analysis_window_validates_end():
    AnalysisWindow(5, 5)
    with pytest.raises(AssertionError):
        AnalysisWindow(5, 4)


@pytest.mark.parametrize(
    "analysis_unit", [AnalysisUnit.CLIENT, AnalysisUnit.PROFILE_GROUP]
)
def test_query_not_detectably_malformed(analysis_unit: AnalysisUnit):
    exp = Experiment("slug", "2019-01-01", 8, analysis_unit=analysis_unit)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl,
        enrollments_query_type=EnrollmentsQueryType.NORMANDY,
        sample_size=None,
    )

    sql_lint(enrollments_sql)
    assert "sample_id < None" not in enrollments_sql

    if analysis_unit == AnalysisUnit.CLIENT:
        assert "client_id" in enrollments_sql
    elif analysis_unit == AnalysisUnit.PROFILE_GROUP:
        assert "profile_group_id" in enrollments_sql

    metrics_sql = exp.build_metrics_query(
        metric_list=[],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    if analysis_unit == AnalysisUnit.CLIENT:
        assert "client_id" in metrics_sql
    elif analysis_unit == AnalysisUnit.PROFILE_GROUP:
        assert "profile_group_id" in metrics_sql


@pytest.mark.parametrize(
    "analysis_unit", [AnalysisUnit.CLIENT, AnalysisUnit.PROFILE_GROUP]
)
def test_megaquery_not_detectably_malformed(analysis_unit: AnalysisUnit):
    exp = Experiment("slug", "2019-01-01", 8, analysis_unit=analysis_unit)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type=EnrollmentsQueryType.NORMANDY
    )

    sql_lint(enrollments_sql)

    if analysis_unit == AnalysisUnit.CLIENT:
        assert "client_id" in enrollments_sql
    elif analysis_unit == AnalysisUnit.PROFILE_GROUP:
        assert "profile_group_id" in enrollments_sql

    metrics_sql = exp.build_metrics_query(
        metric_list=desktop_metrics,
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    if analysis_unit == AnalysisUnit.CLIENT:
        assert "client_id" in metrics_sql
    elif analysis_unit == AnalysisUnit.PROFILE_GROUP:
        assert "profile_group_id" in metrics_sql


@pytest.mark.parametrize(
    "analysis_unit", [AnalysisUnit.CLIENT, AnalysisUnit.PROFILE_GROUP]
)
def test_segments_megaquery_not_detectably_malformed(
    analysis_unit: AnalysisUnit,
):
    exp = Experiment("slug", "2019-01-01", 8, analysis_unit=analysis_unit)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl,
        segment_list=desktop_segments,
        enrollments_query_type=EnrollmentsQueryType.NORMANDY,
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=desktop_metrics,
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)


def test_app_id_propagates():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    sds = SegmentDataSource(
        name="cool_data_source",
        from_expr="`moz-fx-data-shared-prod`.{dataset}.cool_table",
        default_dataset="org_mozilla_firefox",
    )

    segment = Segment(
        name="cool_segment",
        select_expr="COUNT(*)",
        data_source=sds,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl,
        segment_list=[segment],
        enrollments_query_type=EnrollmentsQueryType.FENIX_FALLBACK,
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=fenix_metrics,
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    assert "org_mozilla_firefox" not in enrollments_sql
    assert "my_cool_app" in enrollments_sql

    sql_lint(metrics_sql)


def test_query_not_detectably_malformed_fenix_fallback():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl,
        enrollments_query_type=EnrollmentsQueryType.FENIX_FALLBACK,
        sample_size=10,
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)


def test_firefox_ios_app_id_propagation():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    sds = SegmentDataSource(
        name="cool_data_source",
        from_expr="`moz-fx-data-shared-prod`.{dataset}.cool_table",
        default_dataset="org_mozilla_ios_firefox",
    )

    segment = Segment(
        name="cool_segment",
        select_expr="COUNT(*)",
        data_source=sds,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl,
        segment_list=[segment],
        enrollments_query_type=EnrollmentsQueryType.GLEAN_EVENT,
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=firefox_ios_metrics,
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    assert "org_mozilla_ios_firefox" not in enrollments_sql
    assert "my_cool_app" in enrollments_sql

    sql_lint(metrics_sql)


def test_firefox_klar_app_id_propagation():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    sds = SegmentDataSource(
        name="cool_data_source",
        from_expr="`moz-fx-data-shared-prod`.{dataset}.cool_table",
        default_dataset="org_mozilla_klar",
    )

    segment = Segment(
        name="cool_segment",
        select_expr="COUNT(*)",
        data_source=sds,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl,
        segment_list=[segment],
        enrollments_query_type=EnrollmentsQueryType.GLEAN_EVENT,
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=klar_android_metrics,
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    assert "org_mozilla_klar" not in enrollments_sql
    assert "my_cool_app" in enrollments_sql

    sql_lint(metrics_sql)


def test_firefox_ios_klar_app_id_propagation():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    sds = SegmentDataSource(
        name="cool_data_source",
        from_expr="`moz-fx-data-shared-prod`.{dataset}.cool_table",
        default_dataset="org_mozilla_ios_klar",
    )

    segment = Segment(
        name="cool_segment",
        select_expr="COUNT(*)",
        data_source=sds,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl,
        segment_list=[segment],
        enrollments_query_type=EnrollmentsQueryType.GLEAN_EVENT,
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=klar_ios_metrics,
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    assert "org_mozilla_ios_klar" not in enrollments_sql
    assert "my_cool_app" in enrollments_sql

    sql_lint(metrics_sql)


def test_enrollments_query_cirrus():
    exp = Experiment("slug", "2019-01-01", 8, app_id="monitor_cirrus")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollment_sql = exp.build_enrollments_query(
        time_limits=tl,
        enrollments_query_type=EnrollmentsQueryType.CIRRUS,
    )

    sql_lint(enrollment_sql)

    assert "exposures" in enrollment_sql
    assert 'mozfun.map.get_key(e.extra, "user_id")' in enrollment_sql
    assert "cirrus_events" in enrollment_sql
    assert 'mozfun.map.get_key(event.extra, "user_id")' in enrollment_sql


def test_exposure_query():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollment_sql = exp.build_enrollments_query(
        time_limits=tl,
        enrollments_query_type=EnrollmentsQueryType.GLEAN_EVENT,
    )

    sql_lint(enrollment_sql)

    assert "exposures" in enrollment_sql


def test_exposure_signal_query():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollment_sql = exp.build_enrollments_query(
        time_limits=tl,
        enrollments_query_type=EnrollmentsQueryType.GLEAN_EVENT,
        exposure_signal=ExposureSignal(
            name="exposures",
            data_source=ConfigLoader.get_data_source("baseline", "fenix"),
            select_expr="metrics.counter.events_total_uri_count > 0",
            friendly_name="URI visited exposure",
            description="Exposed when URI visited",
        ),
    )

    sql_lint(enrollment_sql)

    assert "exposures" in enrollment_sql
    assert "metrics.counter.events_total_uri_count > 0" in enrollment_sql


def test_exposure_signal_query_custom_windows():
    exp = Experiment("slug", "2019-01-01", 8, app_id="my_cool_app")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollment_sql = exp.build_enrollments_query(
        time_limits=tl,
        enrollments_query_type=EnrollmentsQueryType.GLEAN_EVENT,
        exposure_signal=ExposureSignal(
            name="exposures",
            data_source=ConfigLoader.get_data_source("baseline", "fenix"),
            select_expr="metrics.counter.events_total_uri_count > 0",
            friendly_name="URI visited exposure",
            description="Exposed when URI visited",
            window_start=1,
            window_end=3,
        ),
    )

    sql_lint(enrollment_sql)

    assert "exposures" in enrollment_sql
    assert "metrics.counter.events_total_uri_count > 0" in enrollment_sql
    assert "DATE_ADD('2019-01-01', INTERVAL 1 DAY)" in enrollment_sql
    assert "DATE_ADD('2019-01-01', INTERVAL 3 DAY)" in enrollment_sql


def test_metrics_query_based_on_exposure():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type=EnrollmentsQueryType.FENIX_FALLBACK
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=fenix_metrics,
        time_limits=tl,
        enrollments_table="enrollments",
        analysis_basis=AnalysisBasis.EXPOSURES,
    )

    sql_lint(metrics_sql)

    assert "e.exposure_date" in metrics_sql


def test_metrics_query_with_exposure_signal_custom_windows():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type=EnrollmentsQueryType.FENIX_FALLBACK
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=fenix_metrics,
        time_limits=tl,
        enrollments_table="enrollments",
        analysis_basis=AnalysisBasis.EXPOSURES,
        exposure_signal=ExposureSignal(
            name="exposures",
            data_source=ConfigLoader.get_data_source("baseline", "fenix"),
            select_expr="metrics.counter.events_total_uri_count > 0",
            friendly_name="URI visited exposure",
            description="Exposed when URI visited",
            window_start=1,
            window_end=3,
        ),
    )

    sql_lint(metrics_sql)

    assert "DATE_ADD('2019-01-01', INTERVAL 1 DAY)" in metrics_sql
    assert "DATE_ADD('2019-01-01', INTERVAL 3 DAY)" in metrics_sql


def test_metrics_query_with_exposure_signal():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type=EnrollmentsQueryType.FENIX_FALLBACK
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=fenix_metrics,
        time_limits=tl,
        enrollments_table="enrollments",
        analysis_basis=AnalysisBasis.EXPOSURES,
        exposure_signal=ExposureSignal(
            name="exposures",
            data_source=ConfigLoader.get_data_source("baseline", "fenix"),
            select_expr="metrics.counter.events_total_uri_count > 0",
            friendly_name="URI visited exposure",
            description="Exposed when URI visited",
        ),
    )

    sql_lint(metrics_sql)

    assert "DATE_ADD('2019-01-01', INTERVAL 0 DAY)" in metrics_sql
    assert "DATE_ADD('2019-01-08', INTERVAL 0 DAY)" in metrics_sql


def test_resolve_metric_slugs():
    exp = Experiment("slug", "2019-01-01", 8, "org_mozilla_fenix", "fenix")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    metrics_sql = exp.build_metrics_query(
        metric_list=["baseline_ping_count"],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    assert "baseline" in metrics_sql


def test_resolve_invalid_metric_slugs():
    exp = Experiment("slug", "2019-01-01", 8, "org_mozilla_fenix", "fenix")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(
        Exception, match="Could not find definition for metric not_exist"
    ):
        exp.build_metrics_query(
            metric_list=["not_exist"],
            time_limits=tl,
            enrollments_table="enrollments",
        )


def test_resolve_invalid_app_name():
    exp = Experiment("slug", "2019-01-01", 8, "unknown", "unknown")

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(ApplicationNotFound, match="Could not find application unknown"):
        exp.build_metrics_query(
            metric_list=["baseline_ping_count"],
            time_limits=tl,
            enrollments_table="enrollments",
        )


def test_resolve_missing_column_names():
    test_ds = DataSource(
        name="test_ds",
        from_expr="SELECT client_id FROM test_set",
        client_id_column=None,
        submission_date_column=None,
    )

    test_metric = Metric(name="test_metric", data_source=test_ds, select_expr="")

    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2022-01-01",
        last_date_full_data="2022-01-13",
        analysis_start_days=0,
        analysis_length_dates=7,
    )

    test_exp = Experiment(experiment_slug="test_exp", start_date="2022-01-01")

    metric_sql = test_exp.build_metrics_query(
        metric_list=[test_metric], time_limits=tl, enrollments_table="test_table"
    )

    assert "None" not in metric_sql


def test_enrollments_query_explicit_client_id():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type=EnrollmentsQueryType.NORMANDY
    )

    sql_lint(enrollments_sql)

    expected = """
    WITH raw_enrollments AS (
SELECT
    e.client_id,
    `mozfun.map.get_key`(e.event_map_values, 'branch')
        AS branch,
    MIN(e.submission_date) AS enrollment_date,
    COUNT(e.submission_date) AS num_enrollment_events
FROM
    `moz-fx-data-shared-prod.telemetry.events` e
WHERE
    e.event_category = 'normandy'
    AND e.event_method = 'enroll'
    AND e.submission_date
        BETWEEN '2019-01-01' AND '2019-01-08'
    AND e.event_string_value = 'slug'
    AND e.sample_id < 100
GROUP BY e.client_id, branch
    ),
    segmented_enrollments AS (
SELECT
    raw_enrollments.*,

FROM raw_enrollments

),
    exposures AS (
SELECT
    e.client_id,
    e.branch,
    min(e.submission_date) AS exposure_date,
    COUNT(e.submission_date) AS num_exposure_events
FROM raw_enrollments re
LEFT JOIN (
    SELECT
        client_id,
        `mozfun.map.get_key`(event_map_values, 'branchSlug') AS branch,
        submission_date
    FROM
        `moz-fx-data-shared-prod.telemetry.events`
    WHERE
        event_category = 'normandy'
        AND (event_method = 'exposure' OR event_method = 'expose')
        AND submission_date
            BETWEEN '2019-01-01' AND '2019-01-08'
        AND event_string_value = 'slug'
) e
ON re.client_id = e.client_id AND
    re.branch = e.branch AND
    e.submission_date >= re.enrollment_date
GROUP BY e.client_id, e.branch
    )

    SELECT
        se.*,
        e.* EXCEPT (client_id, branch)
    FROM segmented_enrollments se
    LEFT JOIN exposures e
    USING (client_id, branch)
"""

    assert dedent(enrollments_sql) == expected

    metrics_sql = exp.build_metrics_query(
        metric_list=[
            metric for metric in desktop_metrics if metric.name == "active_hours"
        ],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)


def test_metrics_query_explicit_client_id():
    exp = Experiment("slug", "2019-01-01", 8)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type=EnrollmentsQueryType.NORMANDY
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[
            metric for metric in desktop_metrics if metric.name == "active_hours"
        ],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    expected = """
WITH analysis_windows AS (
    (SELECT 0 AS analysis_window_start, 6 AS analysis_window_end)
UNION ALL
(SELECT 7 AS analysis_window_start, 13 AS analysis_window_end)
UNION ALL
(SELECT 14 AS analysis_window_start, 20 AS analysis_window_end)
UNION ALL
(SELECT 21 AS analysis_window_start, 27 AS analysis_window_end)
UNION ALL
(SELECT 28 AS analysis_window_start, 34 AS analysis_window_end)
UNION ALL
(SELECT 35 AS analysis_window_start, 41 AS analysis_window_end)
UNION ALL
(SELECT 42 AS analysis_window_start, 48 AS analysis_window_end)
),
raw_enrollments AS (
    -- needed by "exposures" sub query
    SELECT
        e.*,
        aw.*
    FROM `enrollments` e
    CROSS JOIN analysis_windows aw
),
exposures AS (
        SELECT
            *
        FROM raw_enrollments e
    ),
enrollments AS (
    SELECT
        e.* EXCEPT (exposure_date, num_exposure_events),
        x.exposure_date,
        x.num_exposure_events
    FROM exposures x
        RIGHT JOIN raw_enrollments e
        USING (client_id, branch)
)
SELECT
    enrollments.*,
    ds_0.active_hours
FROM enrollments
    LEFT JOIN (
    SELECT
    e.client_id,
    e.branch,
    e.analysis_window_start,
    e.analysis_window_end,
    e.num_exposure_events,
    e.exposure_date,
    COALESCE(SUM(active_hours_sum), 0) AS active_hours
FROM enrollments e
    LEFT JOIN mozdata.telemetry.clients_daily ds
        ON ds.client_id = e.client_id
        AND ds.submission_date BETWEEN '2019-01-01' AND '2019-02-25'
        AND ds.submission_date BETWEEN
            DATE_ADD(e.enrollment_date, interval e.analysis_window_start day)
            AND DATE_ADD(e.enrollment_date, interval e.analysis_window_end day)

GROUP BY
    e.client_id,
    e.branch,
    e.num_exposure_events,
    e.exposure_date,
    e.analysis_window_start,
    e.analysis_window_end
    ) ds_0 USING (client_id, branch, analysis_window_start, analysis_window_end)"""

    assert expected == dedent(metrics_sql.rstrip())


def test_enrollments_query_explicit_group_id():
    exp = Experiment("slug", "2019-01-01", 8, analysis_unit=AnalysisUnit.PROFILE_GROUP)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type=EnrollmentsQueryType.NORMANDY
    )

    sql_lint(enrollments_sql)

    expected = """
    WITH raw_enrollments AS (
SELECT
    e.profile_group_id,
    `mozfun.map.get_key`(e.event_map_values, 'branch')
        AS branch,
    MIN(e.submission_date) AS enrollment_date,
    COUNT(e.submission_date) AS num_enrollment_events
FROM
    `moz-fx-data-shared-prod.telemetry.events` e
WHERE
    e.event_category = 'normandy'
    AND e.event_method = 'enroll'
    AND e.submission_date
        BETWEEN '2019-01-01' AND '2019-01-08'
    AND e.event_string_value = 'slug'
    AND e.sample_id < 100
GROUP BY e.profile_group_id, branch
    ),
    segmented_enrollments AS (
SELECT
    raw_enrollments.*,

FROM raw_enrollments

),
    exposures AS (
SELECT
    e.profile_group_id,
    e.branch,
    min(e.submission_date) AS exposure_date,
    COUNT(e.submission_date) AS num_exposure_events
FROM raw_enrollments re
LEFT JOIN (
    SELECT
        profile_group_id,
        `mozfun.map.get_key`(event_map_values, 'branchSlug') AS branch,
        submission_date
    FROM
        `moz-fx-data-shared-prod.telemetry.events`
    WHERE
        event_category = 'normandy'
        AND (event_method = 'exposure' OR event_method = 'expose')
        AND submission_date
            BETWEEN '2019-01-01' AND '2019-01-08'
        AND event_string_value = 'slug'
) e
ON re.profile_group_id = e.profile_group_id AND
    re.branch = e.branch AND
    e.submission_date >= re.enrollment_date
GROUP BY e.profile_group_id, e.branch
    )

    SELECT
        se.*,
        e.* EXCEPT (profile_group_id, branch)
    FROM segmented_enrollments se
    LEFT JOIN exposures e
    USING (profile_group_id, branch)
"""

    assert dedent(enrollments_sql) == expected


def test_metrics_query_explicit_group_id():
    exp = Experiment("slug", "2019-01-01", 8, analysis_unit=AnalysisUnit.PROFILE_GROUP)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    enrollments_sql = exp.build_enrollments_query(
        time_limits=tl, enrollments_query_type=EnrollmentsQueryType.NORMANDY
    )

    sql_lint(enrollments_sql)

    metrics_sql = exp.build_metrics_query(
        metric_list=[
            metric for metric in desktop_metrics if metric.name == "active_hours"
        ],
        time_limits=tl,
        enrollments_table="enrollments",
    )

    sql_lint(metrics_sql)

    expected = """
WITH analysis_windows AS (
    (SELECT 0 AS analysis_window_start, 6 AS analysis_window_end)
UNION ALL
(SELECT 7 AS analysis_window_start, 13 AS analysis_window_end)
UNION ALL
(SELECT 14 AS analysis_window_start, 20 AS analysis_window_end)
UNION ALL
(SELECT 21 AS analysis_window_start, 27 AS analysis_window_end)
UNION ALL
(SELECT 28 AS analysis_window_start, 34 AS analysis_window_end)
UNION ALL
(SELECT 35 AS analysis_window_start, 41 AS analysis_window_end)
UNION ALL
(SELECT 42 AS analysis_window_start, 48 AS analysis_window_end)
),
raw_enrollments AS (
    -- needed by "exposures" sub query
    SELECT
        e.*,
        aw.*
    FROM `enrollments` e
    CROSS JOIN analysis_windows aw
),
exposures AS (
        SELECT
            *
        FROM raw_enrollments e
    ),
enrollments AS (
    SELECT
        e.* EXCEPT (exposure_date, num_exposure_events),
        x.exposure_date,
        x.num_exposure_events
    FROM exposures x
        RIGHT JOIN raw_enrollments e
        USING (profile_group_id, branch)
)
SELECT
    enrollments.*,
    ds_0.active_hours
FROM enrollments
    LEFT JOIN (
    SELECT
    e.profile_group_id,
    e.branch,
    e.analysis_window_start,
    e.analysis_window_end,
    e.num_exposure_events,
    e.exposure_date,
    COALESCE(SUM(active_hours_sum), 0) AS active_hours
FROM enrollments e
    LEFT JOIN mozdata.telemetry.clients_daily ds
        ON ds.profile_group_id = e.profile_group_id
        AND ds.submission_date BETWEEN '2019-01-01' AND '2019-02-25'
        AND ds.submission_date BETWEEN
            DATE_ADD(e.enrollment_date, interval e.analysis_window_start day)
            AND DATE_ADD(e.enrollment_date, interval e.analysis_window_end day)

GROUP BY
    e.profile_group_id,
    e.branch,
    e.num_exposure_events,
    e.exposure_date,
    e.analysis_window_start,
    e.analysis_window_end
    ) ds_0 USING (profile_group_id, branch, analysis_window_start, analysis_window_end)"""  # noqa: E501

    assert expected == dedent(metrics_sql.rstrip())


def test_glean_group_id_incompatible():
    exp = Experiment(
        "slug",
        "2019-01-01",
        8,
        analysis_unit=AnalysisUnit.PROFILE_GROUP,
        app_id="test_app",
    )

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(IncompatibleAnalysisUnit):
        exp.build_enrollments_query(
            time_limits=tl, enrollments_query_type=EnrollmentsQueryType.GLEAN_EVENT
        )


def test_glean_group_id_incompatible_exposures():
    exp = Experiment(
        "slug",
        "2019-01-01",
        8,
        analysis_unit=AnalysisUnit.PROFILE_GROUP,
        app_id="test_app",
    )

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(IncompatibleAnalysisUnit):
        exp._build_exposure_query(
            time_limits=tl, exposure_query_type=EnrollmentsQueryType.GLEAN_EVENT
        )


def test_glean_missing_app_id():
    exp = Experiment("slug", "2019-01-01", 8, analysis_unit=AnalysisUnit.PROFILE_GROUP)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(
        ValueError, match="App ID must be defined for building Glean enrollments query"
    ):
        exp.build_enrollments_query(
            time_limits=tl, enrollments_query_type=EnrollmentsQueryType.GLEAN_EVENT
        )


def test_glean_exposures_missing_app_id():
    exp = Experiment("slug", "2019-01-01", 8, analysis_unit=AnalysisUnit.PROFILE_GROUP)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(
        ValueError, match="App ID must be defined for building Glean exposures query"
    ):
        exp._build_exposure_query(
            time_limits=tl, exposure_query_type=EnrollmentsQueryType.GLEAN_EVENT
        )


def test_cirrus_group_id_incompatible():
    exp = Experiment(
        "slug",
        "2019-01-01",
        8,
        analysis_unit=AnalysisUnit.PROFILE_GROUP,
        app_id="test_app",
    )

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(IncompatibleAnalysisUnit):
        exp.build_enrollments_query(
            time_limits=tl, enrollments_query_type=EnrollmentsQueryType.CIRRUS
        )


def test_cirrus_group_id_incompatible_exposures():
    exp = Experiment(
        "slug",
        "2019-01-01",
        8,
        analysis_unit=AnalysisUnit.PROFILE_GROUP,
        app_id="test_app",
    )

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(IncompatibleAnalysisUnit):
        exp._build_exposure_query(
            time_limits=tl, exposure_query_type=EnrollmentsQueryType.CIRRUS
        )


def test_cirrus_missing_app_id():
    exp = Experiment("slug", "2019-01-01", 8, analysis_unit=AnalysisUnit.PROFILE_GROUP)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(
        ValueError, match="App ID must be defined for building Cirrus enrollments query"
    ):
        exp.build_enrollments_query(
            time_limits=tl, enrollments_query_type=EnrollmentsQueryType.CIRRUS
        )


def test_cirrus_missing_app_id_exposures():
    exp = Experiment("slug", "2019-01-01", 8, analysis_unit=AnalysisUnit.PROFILE_GROUP)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(
        ValueError, match="App ID must be defined for building Cirrus exposures query"
    ):
        exp._build_exposure_query(
            time_limits=tl, exposure_query_type=EnrollmentsQueryType.CIRRUS
        )


def test_fenix_group_id_incompatible():
    exp = Experiment("slug", "2019-01-01", 8, analysis_unit=AnalysisUnit.PROFILE_GROUP)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(IncompatibleAnalysisUnit):
        exp.build_enrollments_query(
            time_limits=tl, enrollments_query_type=EnrollmentsQueryType.FENIX_FALLBACK
        )


def test_fenix_group_id_incompatible_exposures():
    exp = Experiment("slug", "2019-01-01", 8, analysis_unit=AnalysisUnit.PROFILE_GROUP)

    tl = TimeLimits.for_ts(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-03-01",
        time_series_period="weekly",
        num_dates_enrollment=8,
    )

    with pytest.raises(IncompatibleAnalysisUnit):
        exp._build_exposure_query(
            time_limits=tl, exposure_query_type=EnrollmentsQueryType.FENIX_FALLBACK
        )



import pandas as pd
import pytest
from helpers.cheap_lint import sql_lint  # local helper file
from helpers.config_loader_lists import desktop_metrics, desktop_segments
from mozanalysis.config import ConfigLoader
from mozanalysis.experiment import TimeLimits
from mozanalysis.metrics import DataSource, Metric
from mozanalysis.segments import Segment, SegmentDataSource
from mozanalysis.sizing import HistoricalTarget


class DumbResponse:
    """mock class that is returned by the DumbBQ run_query"""

    @staticmethod
    def to_dataframe():
        """Returns empty dataframe with columns corresponeding to metrics
        tested in test_mixed_metric, test_target_metric_mismatch
        and test_target_metric_mismatch_with_custom"""
        return pd.DataFrame(columns=["active_hours", "baseline_ping_count", "qcdou"])


class DumbBQ:
    def run_query(self, *args, **kwargs):
        """returns a DumbResponse object"""
        return DumbResponse()

    def fully_qualify_table_name(self, *args, **kwargs):
        """just exists so an error isn't thrown"""
        return


def test_mixed_metric():
    # NOTE: no equivalent of this test for targets because only
    # desktop currently has segements defined in metric-hub
    bq_context = DumbBQ()

    ht = HistoricalTarget(
        experiment_name="my_test_name",
        start_date="2021-01-01",
        num_dates_enrollment=2,
        analysis_length=4,
    )

    active_hours = ConfigLoader.get_metric("active_hours", "firefox_desktop")
    baseline_ping_count = ConfigLoader.get_metric(
        "baseline_ping_count", "focus_android"
    )

    allweek_regular_v1 = ConfigLoader.get_segment(
        "allweek_regular_v1", "firefox_desktop"
    )

    with pytest.warns(match="metric_list contains multiple metric-hub apps"):
        _ = ht.get_single_window_data(
            bq_context,
            metric_list=[active_hours, baseline_ping_count],
            target_list=[allweek_regular_v1],
        )


def test_target_metric_mismatch():
    bq_context = DumbBQ()

    ht = HistoricalTarget(
        experiment_name="my_test_name",
        start_date="2021-01-01",
        num_dates_enrollment=2,
        analysis_length=4,
    )

    baseline_ping_count = ConfigLoader.get_metric(
        "baseline_ping_count", "focus_android"
    )

    allweek_regular_v1 = ConfigLoader.get_segment(
        "allweek_regular_v1", "firefox_desktop"
    )

    with pytest.warns(match="metric_list and target_list metric-hub apps do not match"):
        _ = ht.get_single_window_data(
            bq_context,
            metric_list=[baseline_ping_count],
            target_list=[allweek_regular_v1],
        )


def test_target_metric_mismatch_with_custom():
    """Includes a custom metric, so the metric_list check should pass
    but the target_list and metric_list comparision should still fail"""
    bq_context = DumbBQ()

    ht = HistoricalTarget(
        experiment_name="my_test_name",
        start_date="2021-01-01",
        num_dates_enrollment=2,
        analysis_length=4,
    )

    baseline_ping_count = ConfigLoader.get_metric(
        "baseline_ping_count", "focus_android"
    )
    clients_daily = ConfigLoader.get_data_source(
        "search_clients_daily", "firefox_desktop"
    )
    qcdou = Metric(
        name="qcdou",
        data_source=clients_daily,
        select_expr="""COUNTIF(
    active_hours_sum > 0 AND
    scalar_parent_browser_engagement_total_uri_count_normal_and_private_mode_sum > 0
)""",
    )

    allweek_regular_v1 = ConfigLoader.get_segment(
        "allweek_regular_v1", "firefox_desktop"
    )

    with pytest.warns(match="metric_list and target_list metric-hub apps do not match"):
        _ = ht.get_single_window_data(
            bq_context,
            metric_list=[baseline_ping_count, qcdou],
            target_list=[allweek_regular_v1],
        )


def test_multiple_datasource():
    test_target = HistoricalTarget("test_targ", "2022-01-01", 7, 3)

    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2022-01-01",
        last_date_full_data="2022-01-04",
        analysis_start_days=0,
        analysis_length_dates=3,
    )

    test_sds = SegmentDataSource("test_ds", "test_table")
    test_seg = Segment("test_seg", test_sds, "TEST AGG SELECT STATEMENT", "", "")

    target_sql = test_target.build_targets_query(
        time_limits=tl,
        target_list=[
            ConfigLoader.get_segment("new_unique_profiles", "firefox_desktop"),
            test_seg,
        ],
    )

    sql_lint(target_sql)

    assert "ds_1" in target_sql


def test_query_not_detectably_malformed():
    test_target = HistoricalTarget("test_targ", "2022-01-01", 7, 3)

    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2022-01-01",
        last_date_full_data="2022-01-04",
        analysis_start_days=0,
        analysis_length_dates=3,
    )

    target_sql = test_target.build_targets_query(
        time_limits=tl,
        target_list=[
            ConfigLoader.get_segment("new_unique_profiles", "firefox_desktop")
        ],
    )

    sql_lint(target_sql)

    metrics_sql = test_target.build_metrics_query(
        time_limits=tl, metric_list=[], targets_table="targets"
    )

    sql_lint(metrics_sql)


def test_megaquery_not_detectably_malformed():
    test_target = HistoricalTarget("test_targ", "2022-01-01", 7, 3)

    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2022-01-01",
        last_date_full_data="2022-01-04",
        analysis_start_days=0,
        analysis_length_dates=3,
    )

    target_sql = test_target.build_targets_query(
        time_limits=tl,
        target_list=desktop_segments,
    )

    sql_lint(target_sql)

    desktop_metrics_no_experiment_slug = [
        m for m in desktop_metrics if "experiment_slug" not in m.select_expr
    ]

    metrics_sql = test_target.build_metrics_query(
        time_limits=tl,
        metric_list=desktop_metrics_no_experiment_slug,
        targets_table="targets",
    )

    sql_lint(metrics_sql)


def test_custom_query_override_target():
    test_target = HistoricalTarget("test_targ", "2022-01-01", 7, 3)

    test_sds = SegmentDataSource(
        "test_ds",
        "test_table",
    )

    test_seg = Segment("test_seg", test_sds, "TEST AGG SELECT STATEMENT", "", "")

    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2022-01-01",
        last_date_full_data="2022-01-04",
        analysis_start_days=0,
        analysis_length_dates=3,
    )

    custom_query = """
    SELECT * from custom_query_target_table
    """

    target_sql = test_target.build_targets_query(
        time_limits=tl, target_list=[test_seg], custom_targets_query=custom_query
    )

    assert "custom_query_target_table" in target_sql
    assert "TEST AGG SELECT STATEMENT" not in target_sql


def test_resolve_missing_column_names():
    test_ds = DataSource(
        name="test_ds",
        from_expr="SELECT client_id FROM test_set",
        client_id_column=None,
        submission_date_column=None,
    )

    test_metric = Metric(name="test_metric", data_source=test_ds, select_expr="")

    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2022-01-01",
        last_date_full_data="2022-01-13",
        analysis_start_days=0,
        analysis_length_dates=7,
    )

    test_sds = SegmentDataSource(
        name="test_sds",
        from_expr="SELECT * FROM test_sds",
        client_id_column=None,
        submission_date_column=None,
    )

    test_seg = Segment(name="test_seg", data_source=test_sds, select_expr="")

    test_targ = HistoricalTarget(
        experiment_name="test_exp",
        start_date="2022-01-01",
        analysis_length=7,
    )

    targets_sql = test_targ.build_targets_query(time_limits=tl, target_list=[test_seg])

    metrics_sql = test_targ.build_metrics_query(
        time_limits=tl, metric_list=[test_metric], targets_table="test_targ_enrollment"
    )

    assert "None" not in targets_sql
    assert "None" not in metrics_sql



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from dataclasses import dataclass

from metric_config_parser.config import ConfigCollection

from mozanalysis.metrics import DataSource, Metric
from mozanalysis.segments import Segment, SegmentDataSource

METRIC_HUB_JETSTREAM_REPO = "https://github.com/mozilla/metric-hub/tree/main/jetstream"


class ApplicationNotFound(ValueError):
    pass


class MetricNotFound(ValueError):
    pass


class DataSourceNotFound(ValueError):
    pass


class SegmentNotFound(ValueError):
    pass


class SegmentDataSourceNotFound(ValueError):
    pass


class _ConfigLoader:
    """
    Loads config files from an external repository.

    Config objects are converted into mozanalysis native types.
    """

    config_collection: ConfigCollection | None = None

    @property
    def configs(self) -> ConfigCollection:
        configs = getattr(self, "_configs", None)
        if configs:
            return configs

        if self.config_collection is None:
            self.config_collection = ConfigCollection.from_github_repo()
        self._configs = self.config_collection
        return self._configs

    def with_configs_from(
        self, repo_urls: list[str] | None, is_private: bool = False
    ) -> "_ConfigLoader":
        """Load configs from another repository and merge with default configs."""
        if not repo_urls:
            return self

        config_collection = ConfigCollection.from_github_repos(
            repo_urls=repo_urls, is_private=is_private
        )
        self.configs.merge(config_collection)
        return self

    def check_configs_for_app(self, app_name: str) -> bool:
        for definition in self.configs.definitions:
            if app_name == definition.platform:
                return True
        return False

    def get_metric(self, metric_slug: str, app_name: str) -> Metric:
        """Load a metric definition for the given app.

        Returns a :class:`mozanalysis.metrics.Metric` instance.
        """

        metric_definition = self.configs.get_metric_definition(metric_slug, app_name)
        if metric_definition is None:
            if self.check_configs_for_app(app_name):
                raise MetricNotFound(
                    f"Could not find definition for metric {metric_slug} in application {app_name}"  # noqa:E501
                )
            else:
                raise ApplicationNotFound(
                    f"Could not find application {app_name}, so metric {metric_slug} could not be resolved"  # noqa:E501
                )

        return Metric(
            name=metric_definition.name,
            select_expr=self.configs.get_env()
            .from_string(metric_definition.select_expression)
            .render(),
            friendly_name=metric_definition.friendly_name,
            description=metric_definition.description,
            data_source=self.get_data_source(
                metric_definition.data_source.name, app_name
            ),
            bigger_is_better=metric_definition.bigger_is_better,
            app_name=app_name,
        )

    def get_data_source(self, data_source_slug: str, app_name: str) -> DataSource:
        """Load a data source definition for the given app.

        Returns a :class:`mozanalysis.metrics.DataSource` instance.
        """

        data_source_definition = self.configs.get_data_source_definition(
            data_source_slug, app_name
        )
        if data_source_definition is None:
            if self.check_configs_for_app(app_name):
                raise DataSourceNotFound(
                    f"Could not find definition for data source {data_source_slug} in application {app_name}"  # noqa:E501
                )
            else:
                raise ApplicationNotFound(
                    f"Could not find application {app_name}, so data source {data_source_slug} could not be resolved"  # noqa:E501
                )

        return DataSource.from_mcp_data_source(data_source_definition, app_name)

    def get_segment(self, segment_slug: str, app_name: str) -> Segment:
        """Load a segment definition for the given app.

        Returns a :class:`mozanalysis.segments.Segment` instance.
        """

        segment_definition = self.configs.get_segment_definition(segment_slug, app_name)
        if segment_definition is None:
            if self.check_configs_for_app(app_name):
                raise SegmentNotFound(
                    f"Could not find definition for segment {segment_slug} in application {app_name}"  # noqa:E501
                )
            else:
                raise ApplicationNotFound(
                    f"Could not find application {app_name}, so segment {segment_slug} could not be resolved"  # noqa:E501
                )

        return Segment(
            name=segment_definition.name,
            data_source=self.get_segment_data_source(
                segment_definition.data_source.name, app_name
            ),
            select_expr=self.configs.get_env()
            .from_string(segment_definition.select_expression)
            .render(),
            friendly_name=segment_definition.friendly_name,
            description=segment_definition.description,
            app_name=app_name,
        )

    def get_segment_data_source(
        self, data_source_slug: str, app_name: str
    ) -> SegmentDataSource:
        """Load a segment data source definition for the given app.

        Returns a :class:`mozanalysis.segments.SegmentDataSource` instance.
        """

        data_source_definition = self.configs.get_segment_data_source_definition(
            data_source_slug, app_name
        )
        if data_source_definition is None:
            if self.check_configs_for_app(app_name):
                raise SegmentDataSourceNotFound(
                    f"Could not find definition for segment data source {data_source_slug} in application {app_name}"  # noqa:E501
                )
            else:
                raise ApplicationNotFound(
                    f"Could not find application {app_name}, so segment data source {data_source_slug} could not be resolved"  # noqa:E501
                )

        return SegmentDataSource(
            name=data_source_definition.name,
            from_expr=data_source_definition.from_expression,
            window_start=data_source_definition.window_start,
            window_end=data_source_definition.window_end,
            client_id_column=data_source_definition.client_id_column,
            submission_date_column=data_source_definition.submission_date_column,
            default_dataset=data_source_definition.default_dataset,
            app_name=app_name,
        )

    def get_outcome_metric(
        self, metric_slug: str, outcome_slug: str, app_name: str
    ) -> Metric:
        """Load a metric definition from an outcome defined for the given app.

        Parametrized metrics are not supported, since they may not be defined outside
        of an experiment.

        Returns a :class:`mozanalysis.metrics.Metric` instance.
        """
        from mozanalysis.metrics import Metric

        if not self.configs.outcomes:
            self.with_configs_from([METRIC_HUB_JETSTREAM_REPO])

        outcome_spec = self.configs.spec_for_outcome(
            slug=outcome_slug, platform=app_name
        )
        if not outcome_spec:
            raise Exception(f"Could not find definition for outcome {outcome_slug}")

        metric_definition = outcome_spec.metrics.get(metric_slug)
        if metric_definition is None:
            raise Exception(
                f"Could not find definition for metric {metric_slug}"
                + f" in outcome {outcome_slug}"
            )

        @dataclass
        class MinimalConfiguration:
            app_name: str

        conf = MinimalConfiguration(app_name)
        summaries = metric_definition.resolve(outcome_spec, conf, self.configs)
        metric = summaries[0].metric

        if metric.data_source is None:
            raise ValueError(f"Unable to resolve DataSource for Metric {metric.name}")

        return Metric(
            name=metric.name,
            select_expr=metric.select_expression,
            friendly_name=metric.friendly_name,
            description=metric.description,
            data_source=DataSource.from_mcp_data_source(metric.data_source, app_name),
            bigger_is_better=metric.bigger_is_better,
        )

    def get_outcome_data_source(
        self, data_source_slug: str, outcome_slug: str, app_name: str
    ) -> DataSource:
        """Load a data source definition from an outcome defined for the given app.

        Returns a :class:`mozanalysis.metrics.DataSource` instance.
        """
        from mozanalysis.metrics import DataSource

        if not self.configs.outcomes:
            self.with_configs_from([METRIC_HUB_JETSTREAM_REPO])

        outcome_spec = self.configs.spec_for_outcome(
            slug=outcome_slug, platform=app_name
        )
        if not outcome_spec:
            raise Exception(f"Could not find definition for outcome {outcome_slug}")

        data_source_definition = outcome_spec.data_sources.definitions.get(
            data_source_slug
        )
        if data_source_definition is None:
            raise Exception(
                f"Could not find definition for data source {data_source_slug}"
                + f" in outcome {outcome_slug}"
            )

        return DataSource.from_mcp_data_source(
            data_source_definition, app_name=app_name
        )


ConfigLoader = _ConfigLoader()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pytest
from helpers.cheap_lint import sql_lint  # local helper file
from mozanalysis.config import (
    ApplicationNotFound,
    ConfigLoader,
    DataSourceNotFound,
    MetricNotFound,
    SegmentDataSourceNotFound,
    SegmentNotFound,
)
from mozanalysis.metrics import DataSource, Metric
from mozanalysis.segments import Segment, SegmentDataSource


def test_get_metric():
    m = ConfigLoader.get_metric("active_hours", "firefox_desktop")
    assert isinstance(m, Metric)
    assert m.name == "active_hours"
    assert m.friendly_name
    assert m.description
    assert m.data_source
    sql_lint(m.select_expr)


def test_unknown_metric_fails():
    with pytest.raises(
        MetricNotFound,
        match="Could not find definition for metric fake_metric in application firefox_desktop",  # noqa: E501
    ):
        ConfigLoader.get_metric("fake_metric", "firefox_desktop")


def test_metric_unknown_app_fails():
    with pytest.raises(
        ApplicationNotFound,
        match="Could not find application fake_app, so metric fake_metric could not be resolved",  # noqa: E501
    ):
        ConfigLoader.get_metric("fake_metric", "fake_app")


def test_get_data_source():
    d = ConfigLoader.get_data_source("clients_daily", "firefox_desktop")
    assert isinstance(d, DataSource)
    assert d.name == "clients_daily"
    assert d.client_id_column == "client_id"
    assert d.submission_date_column == "submission_date"
    sql_lint(d.from_expr_for("dataset"))


def test_unknown_data_source_fails():
    with pytest.raises(
        DataSourceNotFound,
        match="Could not find definition for data source fake_data_source in application firefox_desktop",  # noqa: E501
    ):
        ConfigLoader.get_data_source("fake_data_source", "firefox_desktop")


def test_data_source_unknown_app_fails():
    with pytest.raises(
        ApplicationNotFound,
        match="Could not find application fake_app, so data source fake_data_source could not be resolved",  # noqa: E501
    ):
        ConfigLoader.get_data_source("fake_data_source", "fake_app")


def test_get_segment():
    s = ConfigLoader.get_segment("suggest", "firefox_desktop")
    assert isinstance(s, Segment)
    assert s.name == "suggest"
    assert s.data_source
    sql_lint(s.select_expr)


def test_unknown_segment_fails():
    with pytest.raises(
        SegmentNotFound,
        match="Could not find definition for segment fake_segment in application firefox_desktop",  # noqa: E501
    ):
        ConfigLoader.get_segment("fake_segment", "firefox_desktop")


def test_segment_unknown_app_fails():
    with pytest.raises(
        ApplicationNotFound,
        match="Could not find application fake_app, so segment fake_segment could not be resolved",  # noqa: E501
    ):
        ConfigLoader.get_segment("fake_segment", "fake_app")


def test_get_segment_data_source():
    s = ConfigLoader.get_segment_data_source("clients_last_seen", "firefox_desktop")
    assert isinstance(s, SegmentDataSource)
    assert s.name == "clients_last_seen"
    assert s.client_id_column == "client_id"
    assert s.submission_date_column == "submission_date"
    assert s.window_start == 0
    assert s.window_end == 0
    sql_lint(s.from_expr_for("dataset"))


def test_unknown_segment_data_source_fails():
    with pytest.raises(
        SegmentDataSourceNotFound,
        match="Could not find definition for segment data source fake_segment_data_source in application firefox_desktop",  # noqa: E501
    ):
        ConfigLoader.get_segment_data_source(
            "fake_segment_data_source", "firefox_desktop"
        )


def test_segmentdatasource_unknown_app_fails():
    with pytest.raises(
        ApplicationNotFound,
        match="Could not find application fake_app, so segment data source fake_segment_data_source could not be resolved",  # noqa: E501
    ):
        ConfigLoader.get_segment_data_source("fake_segment_data_source", "fake_app")


def test_get_outcome_metric_outcome_data_source():
    m = ConfigLoader.get_outcome_metric(
        "cert_error_page_loaded", "networking", "firefox_desktop"
    )
    assert isinstance(m, Metric)
    assert m.name == "cert_error_page_loaded"
    assert m.friendly_name
    assert m.data_source
    sql_lint(m.select_expr)


def test_get_outcome_metric_general_data_source():
    m = ConfigLoader.get_outcome_metric(
        "urlbar_search_count", "firefox_suggest", "firefox_desktop"
    )
    assert isinstance(m, Metric)
    assert m.name == "urlbar_search_count"
    assert m.friendly_name
    assert m.description
    assert m.data_source
    sql_lint(m.select_expr)


def test_get_parametrized_outcome_metric_fails():
    m = ConfigLoader.get_outcome_metric(
        "spotlight_impressions", "spotlight_engagement", "firefox_desktop"
    )
    assert isinstance(m, Metric)
    assert m.name == "spotlight_impressions"
    assert m.friendly_name
    assert m.description
    assert m.data_source
    sql_lint(m.select_expr)


def test_outcome_unknown_metric_fails():
    with pytest.raises(
        Exception,
        match="Could not find definition for metric fake_metric in outcome networking",
    ):
        ConfigLoader.get_outcome_metric("fake_metric", "networking", "firefox_desktop")


def test_unknown_outcome_metric_fails():
    with pytest.raises(
        Exception, match="Could not find definition for outcome fake_outcome"
    ):
        ConfigLoader.get_outcome_metric(
            "fake_metric", "fake_outcome", "firefox_desktop"
        )


def test_get_outcome_data_source():
    d = ConfigLoader.get_outcome_data_source(
        "events_certerror", "networking", "firefox_desktop"
    )
    assert isinstance(d, DataSource)
    assert d.name == "events_certerror"
    assert d.submission_date_column == "submission_date"
    sql_lint(d.from_expr_for("dataset"))


def test_outcome_unknown_data_source():
    with pytest.raises(
        Exception,
        match=(
            "Could not find definition for data source fake_data_source"
            + " in outcome networking"
        ),
    ):
        ConfigLoader.get_outcome_data_source(
            "fake_data_source", "networking", "firefox_desktop"
        )


def test_unknown_outcome_data_source():
    with pytest.raises(
        Exception, match="Could not find definition for outcome fake_outcome"
    ):
        ConfigLoader.get_outcome_data_source(
            "fake_data_source", "fake_outcome", "firefox_desktop"
        )



import logging
import re
from copy import deepcopy
from itertools import product

import mozanalysis.frequentist_stats.bootstrap as mafsb
import mozanalysis.frequentist_stats.linear_models as mafslm
import mozanalysis.frequentist_stats.linear_models.functions as func
import mozanalysis.types as types
import numpy as np
import pandas as pd
import polars as pl
import pytest
import statsmodels.formula.api as smf
from mozanalysis.frequentist_stats.linear_models.functions import (
    FailedToFitModel,
    MissingModelParameter,
)
from polars.testing import assert_frame_equal as pl_assert_frame_equal
from statsmodels.stats.weightstats import CompareMeans

from .helpers import test_model, test_model_covariate

logger = logging.getLogger(__name__)


def test__stringify_alpha():
    for bad_alphas in [-1, 0, 1, 2]:
        with pytest.raises(ValueError, match=r"alpha must be in \(0.002,1\)"):
            func._stringify_alpha(bad_alphas)

    alpha = 0.1
    low, high = func._stringify_alpha(alpha)
    assert low == "0.050"
    assert high == "0.950"

    alpha = 0.05
    low, high = func._stringify_alpha(alpha)
    assert low == "0.025"
    assert high == "0.975"

    alpha = 0.01
    low, high = func._stringify_alpha(alpha)
    assert low == "0.005"
    assert high == "0.995"


def test_summarize_one_branch():
    test_data = pd.Series(range(100))
    alphas = [0.05]
    actuals = mafslm.summarize_one_branch(test_data, alphas)

    mean = 49.5
    low, high = 43.74349, 55.25650

    assert np.isclose(actuals["0.5"], mean)
    assert np.isclose(actuals["mean"], mean)
    assert np.isclose(actuals["0.025"], low)
    assert np.isclose(actuals["0.975"], high)

    index_values = actuals.index.values
    index_values.sort()
    assert list(index_values) == ["0.025", "0.5", "0.975", "mean"]


def test_summarize_univariate():
    control_branch_data = pd.Series(range(100))
    treatment_branch_data = pd.Series(range(100, 200))
    test_data = pd.concat([control_branch_data, treatment_branch_data])
    branches = pd.Series([*(["control"] * 100), *(["treatment"] * 100)])
    branch_list = ["control", "treatment"]

    result = mafslm.summarize_univariate(test_data, branches, [0.05])

    assert sorted(result.keys()) == branch_list
    for branch in branch_list:
        index_values = result[branch].index.values
        index_values.sort()
        assert list(index_values) == ["0.025", "0.5", "0.975", "mean"]

    # control
    ## validate against theoretical values
    mean = 49.5
    low, high = 43.74349, 55.25650
    assert np.isclose(result["control"]["mean"], mean)
    assert np.isclose(result["control"]["0.5"], mean)
    assert np.isclose(result["control"]["0.025"], low)
    assert np.isclose(result["control"]["0.975"], high)

    ## cross-validate against existing bootstrap implementation
    bootstrap_result = mafsb.bootstrap_one_branch(control_branch_data)
    assert np.isclose(result["control"]["mean"], bootstrap_result["mean"], atol=0.5)
    assert np.isclose(result["control"]["0.5"], bootstrap_result["0.5"], atol=0.5)
    assert np.isclose(result["control"]["0.025"], bootstrap_result["0.025"], atol=0.5)
    assert np.isclose(result["control"]["0.975"], bootstrap_result["0.975"], atol=0.5)

    # treatment
    ## validate against theoretical values
    mean = 149.5
    low, high = 143.74349, 155.25650
    assert np.isclose(result["treatment"]["mean"], mean)
    assert np.isclose(result["treatment"]["0.5"], mean)
    assert np.isclose(result["treatment"]["0.025"], low)
    assert np.isclose(result["treatment"]["0.975"], high)

    ## cross-validate against existing bootstrap implementation
    bootstrap_result = mafsb.bootstrap_one_branch(treatment_branch_data)
    assert np.isclose(result["treatment"]["mean"], bootstrap_result["mean"], atol=0.5)
    assert np.isclose(result["treatment"]["0.5"], bootstrap_result["0.5"], atol=0.5)
    assert np.isclose(result["treatment"]["0.025"], bootstrap_result["0.025"], atol=0.5)
    assert np.isclose(result["treatment"]["0.975"], bootstrap_result["0.975"], atol=0.5)


def test_make_formula():
    expected = "search_count ~ C(branch, Treatment(reference='control'))"
    actual = mafslm.make_formula("search_count", "control")

    assert expected == actual

    expected = "days_of_use ~ C(branch, Treatment(reference='treatment-a'))"
    actual = mafslm.make_formula("days_of_use", "treatment-a")

    assert expected == actual

    expected = "active_hours ~ C(branch, Treatment(reference='treatment-a')) + active_hours_pre"  # noqa: E501
    actual = mafslm.make_formula("active_hours", "treatment-a", "active_hours_pre")

    assert expected == actual

    for bad_target in ["search~count", "search(count", "search)count", "search_count'"]:
        with pytest.raises(
            ValueError, match=r"Target variable .* contains invalid character"
        ):
            mafslm.make_formula(bad_target, "control")

    for bad_branch in ["search~count", "search(count", "search)count", "search_count'"]:
        with pytest.raises(
            ValueError, match=r"Reference branch .* contains invalid character"
        ):
            mafslm.make_formula("search_count", bad_branch)

    for bad_covariate in [
        "search~count",
        "search(count",
        "search)count",
        "search_count'",
    ]:
        with pytest.raises(
            ValueError, match=r"Covariate .* contains invalid character"
        ):
            mafslm.make_formula("search_count", "control", bad_covariate)


def test__make_joint_output():
    out = func._make_joint_output([0.01, 0.05], types.Uplift.RELATIVE)

    expected_keys = [
        ("rel_uplift", "exp"),
        ("rel_uplift", "0.5"),
        ("rel_uplift", "0.005"),
        ("rel_uplift", "0.995"),
        ("rel_uplift", "0.025"),
        ("rel_uplift", "0.975"),
    ]

    for key in expected_keys:
        assert key in out.index

    assert len(out.index) == len(expected_keys)


def test__extract_absolute_uplifts():
    bootstrap_results = mafsb.compare_branches(
        test_model.model_df,
        test_model.target,
        ref_branch_label=test_model.ref_branch,
    )

    # control vs treatment-a
    out = func._extract_absolute_uplifts(
        test_model.results, "control", test_model.ref_branch, test_model.alphas
    )

    expected_mean = 50 - 150
    assert np.isclose(out[("abs_uplift", "exp")], expected_mean)
    assert np.isclose(out[("abs_uplift", "0.5")], expected_mean)

    for alpha in test_model.alphas:
        a_str_low, a_str_high = func._stringify_alpha(alpha)
        low, high = CompareMeans.from_data(
            test_model.model_df.loc[
                test_model.model_df.branch == "control", test_model.target
            ],
            test_model.model_df.loc[
                test_model.model_df.branch == test_model.ref_branch, test_model.target
            ],
        ).zconfint_diff(alpha)
        assert np.isclose(out[("abs_uplift", a_str_low)], low, atol=0.1)
        assert np.isclose(out[("abs_uplift", a_str_high)], high, atol=0.1)

    ## cross-validate with existing bootstrap implementation
    boot_df = bootstrap_results["comparative"]["control"]
    for alpha in test_model.alphas:
        a_str_low, a_str_high = func._stringify_alpha(alpha)
        assert np.isclose(
            out[("abs_uplift", a_str_low)], boot_df[("abs_uplift", a_str_low)], atol=1.0
        )
        assert np.isclose(
            out[("abs_uplift", a_str_high)],
            boot_df[("abs_uplift", a_str_high)],
            atol=1.0,
        )
    # treatment-a vs treatment-b
    out = func._extract_absolute_uplifts(
        test_model.results, "treatment-b", test_model.ref_branch, test_model.alphas
    )

    expected_mean = 250 - 150
    assert np.isclose(out[("abs_uplift", "exp")], expected_mean)
    assert np.isclose(out[("abs_uplift", "0.5")], expected_mean)

    for alpha in test_model.alphas:
        a_str_low, a_str_high = func._stringify_alpha(alpha)
        low, high = CompareMeans.from_data(
            test_model.model_df.loc[
                test_model.model_df.branch == "treatment-b", test_model.target
            ],
            test_model.model_df.loc[
                test_model.model_df.branch == test_model.ref_branch, test_model.target
            ],
        ).zconfint_diff(alpha)
        assert np.isclose(out[("abs_uplift", a_str_low)], low, atol=0.1)
        assert np.isclose(out[("abs_uplift", a_str_high)], high, atol=0.1)

    ## cross-validate with existing bootstrap implementation
    boot_df = bootstrap_results["comparative"]["treatment-b"]
    for alpha in test_model.alphas:
        a_str_low, a_str_high = func._stringify_alpha(alpha)
        assert np.isclose(
            out[("abs_uplift", a_str_low)], boot_df[("abs_uplift", a_str_low)], atol=1.0
        )
        assert np.isclose(
            out[("abs_uplift", a_str_high)],
            boot_df[("abs_uplift", a_str_high)],
            atol=1.0,
        )


def test__extract_absolute_uplifts_covariate():
    bootstrap_results = mafsb.compare_branches(
        test_model_covariate.model_df,
        test_model_covariate.target,
        ref_branch_label=test_model_covariate.ref_branch,
    )

    # control vs treatment-a
    out = func._extract_absolute_uplifts(
        test_model_covariate.results,
        "control",
        test_model_covariate.ref_branch,
        test_model_covariate.alphas,
    )

    expected_mean = -0.1
    assert np.isclose(out[("abs_uplift", "exp")], expected_mean, atol=0.05)
    assert np.isclose(out[("abs_uplift", "0.5")], expected_mean, atol=0.05)

    ## compare to bootstrap implementation
    # validate the point estimates are more precise
    boot_df = bootstrap_results["comparative"]["control"]

    ###validate the widths of confidence intervals are smaller with LM approach
    for alpha in test_model_covariate.alphas:
        a_str_low, a_str_high = func._stringify_alpha(alpha)
        ci_width_boot = (
            boot_df[("abs_uplift", a_str_high)] - boot_df[("abs_uplift", a_str_low)]
        )
        ci_width_lm = out[("abs_uplift", a_str_high)] - out[("abs_uplift", a_str_low)]
        assert ci_width_lm < ci_width_boot

    # treatment-a vs treatment-b
    out = func._extract_absolute_uplifts(
        test_model_covariate.results,
        "treatment-b",
        test_model_covariate.ref_branch,
        test_model_covariate.alphas,
    )

    expected_mean = 0.1
    assert np.isclose(out[("abs_uplift", "exp")], expected_mean, atol=0.05)
    assert np.isclose(out[("abs_uplift", "0.5")], expected_mean, atol=0.05)

    boot_df = bootstrap_results["comparative"]["treatment-b"]
    for alpha in test_model_covariate.alphas:
        a_str_low, a_str_high = func._stringify_alpha(alpha)
        ci_width_boot = (
            boot_df[("abs_uplift", a_str_high)] - boot_df[("abs_uplift", a_str_low)]
        )
        ci_width_lm = out[("abs_uplift", a_str_high)] - out[("abs_uplift", a_str_low)]
        assert ci_width_lm < ci_width_boot


def test__extract_relative_uplifts():
    bootstrap_results = mafsb.compare_branches(
        test_model.model_df, test_model.target, ref_branch_label=test_model.ref_branch
    )

    # control vs treatment-a
    out = func._extract_relative_uplifts(
        test_model.results,
        "control",
        test_model.ref_branch,
        test_model.alphas,
        test_model.treatment_branches,
    )

    expected_mean = (50 - 150) / 150
    assert np.isclose(out[("rel_uplift", "exp")], expected_mean, atol=0.01)
    assert np.isclose(out[("rel_uplift", "0.5")], expected_mean, atol=0.01)

    ## cross-validate with existing bootstrap implementation
    boot_df = bootstrap_results["comparative"]["control"]
    for alpha in test_model.alphas:
        a_str_low, a_str_high = func._stringify_alpha(alpha)
        assert np.isclose(
            out[("rel_uplift", a_str_low)],
            boot_df[("rel_uplift", a_str_low)],
            atol=0.01,
        )
        assert np.isclose(
            out[("rel_uplift", a_str_high)],
            boot_df[("rel_uplift", a_str_high)],
            atol=0.01,
        )

    out = func._extract_relative_uplifts(
        test_model.results,
        "treatment-b",
        test_model.ref_branch,
        test_model.alphas,
        test_model.treatment_branches,
    )

    expected_mean = (250 - 150) / 150
    assert np.isclose(out[("rel_uplift", "exp")], expected_mean, atol=0.01)
    assert np.isclose(out[("rel_uplift", "0.5")], expected_mean, atol=0.01)

    ## cross-validate with existing bootstrap implementation
    boot_df = bootstrap_results["comparative"]["treatment-b"]
    for alpha in test_model.alphas:
        a_str_low, a_str_high = func._stringify_alpha(alpha)
        assert np.isclose(
            out[("rel_uplift", a_str_low)],
            boot_df[("rel_uplift", a_str_low)],
            atol=0.01,
        )
        assert np.isclose(
            out[("rel_uplift", a_str_high)],
            boot_df[("rel_uplift", a_str_high)],
            atol=0.01,
        )


def test__extract_relative_uplifts_covariate():
    bootstrap_results = mafsb.compare_branches(
        test_model_covariate.model_df,
        test_model_covariate.target,
        ref_branch_label=test_model_covariate.ref_branch,
    )

    # control vs treatment-a
    out = func._extract_relative_uplifts(
        test_model_covariate.results,
        "control",
        test_model_covariate.ref_branch,
        test_model_covariate.alphas,
        test_model_covariate.treatment_branches,
    )

    expected_mean = -0.1 / 2.1
    assert np.isclose(out[("rel_uplift", "exp")], expected_mean, atol=0.05)
    assert np.isclose(out[("rel_uplift", "0.5")], expected_mean, atol=0.05)

    ## compare to bootstrap implementation
    # validate the point estimates are more precise
    boot_df = bootstrap_results["comparative"]["control"]

    ###validate the widths of confidence intervals are smaller with LM approach
    for alpha in test_model_covariate.alphas:
        a_str_low, a_str_high = func._stringify_alpha(alpha)
        ci_width_boot = (
            boot_df[("rel_uplift", a_str_high)] - boot_df[("rel_uplift", a_str_low)]
        )
        ci_width_lm = out[("rel_uplift", a_str_high)] - out[("rel_uplift", a_str_low)]
        assert ci_width_lm < ci_width_boot

    # treatment-a vs treatment-b
    out = func._extract_relative_uplifts(
        test_model_covariate.results,
        "treatment-b",
        test_model_covariate.ref_branch,
        test_model_covariate.alphas,
        test_model_covariate.treatment_branches,
    )

    expected_mean = 0.1 / 2.1
    assert np.isclose(out[("rel_uplift", "exp")], expected_mean, atol=0.05)
    assert np.isclose(out[("rel_uplift", "0.5")], expected_mean, atol=0.05)

    boot_df = bootstrap_results["comparative"]["treatment-b"]
    for alpha in test_model_covariate.alphas:
        a_str_low, a_str_high = func._stringify_alpha(alpha)
        ci_width_boot = (
            boot_df[("rel_uplift", a_str_high)] - boot_df[("rel_uplift", a_str_low)]
        )
        ci_width_lm = out[("rel_uplift", a_str_high)] - out[("rel_uplift", a_str_low)]
        assert ci_width_lm < ci_width_boot


@pytest.mark.parametrize("deallocate", [True, False])
def test_summarize_joint(deallocate):
    """Validates the structure of the comparative results object,
    tests for accuracy of reported values are found above, in the
    test__extract_<>_uplifts tests"""

    actual = mafslm.summarize_joint(
        test_model.model_df,
        test_model.target,
        test_model.alphas,
        ref_branch_label="treatment-a",
        deallocate_aggressively=deallocate,
    )

    expected_index = [
        ("abs_uplift", "0.5"),
        ("abs_uplift", "exp"),
        ("abs_uplift", "0.005"),
        ("abs_uplift", "0.995"),
        ("abs_uplift", "0.025"),
        ("abs_uplift", "0.975"),
        ("rel_uplift", "0.5"),
        ("rel_uplift", "exp"),
        ("rel_uplift", "0.005"),
        ("rel_uplift", "0.995"),
        ("rel_uplift", "0.025"),
        ("rel_uplift", "0.975"),
    ]

    for branch in ["control", "treatment-b"]:
        for key in expected_index:
            assert key in actual[branch].index
        assert len(actual[branch].index) == len(expected_index)


@pytest.mark.parametrize("copy", [True, False])
def test_prepare_df_for_modeling(copy):
    # test removal of nulls
    y = [1, 2, 3, 4, None, np.nan]
    branch = ["control", "control", "treatment", "treatment", "control", "treatment"]
    df_in = pd.DataFrame({"y": y, "branch": branch})
    df_actual = mafslm.prepare_df_for_modeling(df_in, "y", copy=copy)
    df_expected = pd.DataFrame({"y": [1.0, 2.0, 3.0, 4.0], "branch": branch[:4]})
    pd.testing.assert_frame_equal(df_actual, df_expected)
    # null removal does not drop rows in place
    # pd.testing.assert_frame_equal(df_in, df_expected)

    # test removal of nulls with covariate
    y = [1, 2, 3, 4, None, np.nan]
    y2 = [1, 2, None, np.nan, 3, 4]
    branch = ["control", "control", "treatment", "treatment", "control", "treatment"]
    df_in = pd.DataFrame({"y": y, "branch": branch, "y2": y2})
    df_actual = mafslm.prepare_df_for_modeling(
        df_in, "y", covariate_col="y2", copy=copy
    )
    df_expected = pd.DataFrame(
        {"y": [1.0, 2.0], "branch": branch[:2], "y2": [1.0, 2.0]}
    )
    pd.testing.assert_frame_equal(df_actual, df_expected)

    # test thresholding ints
    y = list(range(100))
    branch = ["control"] * 100
    df_in = pd.DataFrame({"y": y, "branch": branch})
    df_actual = mafslm.prepare_df_for_modeling(
        df_in, "y", threshold_quantile=0.95, copy=copy
    )
    df_expected = pd.DataFrame(
        {
            "y": [int(_y) for _y in y[:-5] + [int(np.ceil(np.quantile(y, 0.95)))] * 5],
            "branch": branch,
        }
    )
    pd.testing.assert_frame_equal(df_actual, df_expected)
    if not copy:
        pd.testing.assert_frame_equal(df_in, df_expected)

    # test thresholding floats
    y = [float(_y) for _y in (range(100))]
    branch = ["control"] * 100
    df_in = pd.DataFrame({"y": y, "branch": branch})
    df_actual = mafslm.prepare_df_for_modeling(
        df_in, "y", threshold_quantile=0.95, copy=copy
    )
    df_expected = pd.DataFrame(
        {
            "y": [float(_y) for _y in y[:-5] + [np.quantile(y, 0.95)] * 5],
            "branch": branch,
        }
    )
    pd.testing.assert_frame_equal(df_actual, df_expected)
    if not copy:
        pd.testing.assert_frame_equal(df_in, df_expected)

    # test covariate & thresholding
    y2 = [float(_y) for _y in range(100, 200)]
    branch = ["control"] * 100
    df_in = pd.DataFrame({"y": y, "branch": branch, "y2": y2})
    df_actual = mafslm.prepare_df_for_modeling(
        df_in, "y", covariate_col="y2", threshold_quantile=0.95, copy=copy
    )
    df_expected = pd.DataFrame(
        {
            "y": [float(_y) for _y in y[:-5] + [np.quantile(y, 0.95)] * 5],
            "branch": branch,
            "y2": [float(_y) for _y in y2[:-5] + [np.quantile(y2, 0.95)] * 5],
        }
    )
    pd.testing.assert_frame_equal(df_actual, df_expected)
    if not copy:
        pd.testing.assert_frame_equal(df_in, df_expected)

    # test behavior of clipping with integer dtypes
    y = list(range(100))
    branch = ["control"] * 100
    df_in = pd.DataFrame({"y": y, "branch": branch})
    df_in["y"] = df_in.y.astype(pd.Int64Dtype())
    df_actual = mafslm.prepare_df_for_modeling(
        df_in, "y", threshold_quantile=0.948, copy=copy
    )
    df_expected = pd.DataFrame(
        {
            "y": [int(_y) for _y in y[:-5] + [int(np.ceil(np.quantile(y, 0.948)))] * 5],
            "branch": branch,
        }
    )
    df_expected["y"] = df_expected.y.astype(pd.Int64Dtype())
    pd.testing.assert_frame_equal(df_actual, df_expected)


@pytest.mark.parametrize("interactive", [True, False])
@pytest.mark.parametrize("deallocate", [True, False])
def test_compare_branches_lm(interactive, deallocate):
    """This is an integration type test, testing only that the
    format of the output object is correct. Functionality testing
    of specific elements of the object is covered by other tests"""
    branches = ["control", "treatment-a", "treatment-b"]

    out = mafslm.compare_branches_lm(
        test_model.model_df,
        test_model.target,
        "treatment-a",
        alphas=test_model.alphas,
        interactive=interactive,
        deallocate_aggressively=deallocate,
    )

    assert list(out.keys()) == ["individual", "comparative"]

    # assert sorted(list(out['individual'].keys())) == branches

    for branch in branches:
        assert sorted(out["individual"][branch].index) == [
            "0.005",
            "0.025",
            "0.5",
            "0.975",
            "0.995",
            "mean",
        ]

    comparative_branches = ["control", "treatment-b"]

    assert sorted(out["comparative"].keys()) == comparative_branches

    expected_index = [
        ("abs_uplift", "0.005"),
        ("abs_uplift", "0.025"),
        ("abs_uplift", "0.5"),
        ("abs_uplift", "0.975"),
        ("abs_uplift", "0.995"),
        ("abs_uplift", "exp"),
        ("rel_uplift", "0.005"),
        ("rel_uplift", "0.025"),
        ("rel_uplift", "0.5"),
        ("rel_uplift", "0.975"),
        ("rel_uplift", "0.995"),
        ("rel_uplift", "exp"),
    ]

    for branch in comparative_branches:
        assert sorted(out["comparative"][branch].index) == expected_index


def test_compare_branches_lm_fallback(caplog):
    # tests whether the results when called with a missing covariate equal the
    # results when called without a covariate. Also tests that a log was emitted.
    msg = "Covariate fake_covariate not found, falling back to unadjusted inferences"

    assert msg not in caplog.text

    actual = mafslm.compare_branches_lm(
        test_model_covariate.model_df,
        test_model_covariate.target,
        test_model_covariate.ref_branch,
        "fake_covariate",
    )

    expected = mafslm.compare_branches_lm(
        test_model_covariate.model_df,
        test_model_covariate.target,
        test_model_covariate.ref_branch,
    )

    for branch in test_model_covariate.treatment_branches:
        pd.testing.assert_series_equal(
            actual["comparative"][branch], expected["comparative"][branch]
        )

    assert msg in caplog.text


def test_fit_model():
    actual_results = mafslm.fit_model(
        test_model.model_df,
        test_model.target,
        test_model.ref_branch,
        test_model.treatment_branches,
    )

    pd.testing.assert_series_equal(actual_results.params, test_model.results.params)

    alpha = 0.05
    pd.testing.assert_frame_equal(
        actual_results.conf_int(alpha), test_model.results.conf_int(alpha)
    )


def test_fit_model_covariate():
    actual_results = mafslm.fit_model(
        test_model_covariate.model_df,
        test_model_covariate.target,
        test_model_covariate.ref_branch,
        test_model_covariate.treatment_branches,
        test_model_covariate.covariate,
    )

    pd.testing.assert_series_equal(
        actual_results.params, test_model_covariate.results.params
    )

    alpha = 0.05
    pd.testing.assert_frame_equal(
        actual_results.conf_int(alpha), test_model_covariate.results.conf_int(alpha)
    )


def test_fit_model_covariate_robust_to_bad_covariate():
    model_df = test_model_covariate.model_df.copy()
    model_df.loc[:, test_model_covariate.covariate] = [0] * model_df.shape[0]

    expected_results = smf.ols(test_model.formula, model_df).fit()

    with pytest.warns(Warning, match="Unexpectedly fell back to unadjusted inferences"):
        actual_results = mafslm.fit_model(
            model_df,
            test_model_covariate.target,
            test_model_covariate.ref_branch,
            test_model_covariate.treatment_branches,
            test_model_covariate.covariate,
        )

    pd.testing.assert_series_equal(actual_results.params, expected_results.params)

    alpha = 0.05
    pd.testing.assert_frame_equal(
        actual_results.conf_int(alpha), expected_results.conf_int(alpha)
    )


def test_fit_model_covariate_fails_on_bad_data():
    model_df = test_model_covariate.model_df.copy()
    model_df.loc[:, test_model_covariate.target] = [0] * model_df.shape[0]

    with pytest.raises(
        FailedToFitModel,
        match="Failed to fit model for target search_count using covariate search_count_pre",  # noqa: E501
    ):
        mafslm.fit_model(
            model_df,
            test_model_covariate.target,
            test_model_covariate.ref_branch,
            test_model_covariate.treatment_branches,
            test_model_covariate.covariate,
        )


def test_fit_model_covariate_fails_on_bad_branch():
    model_df = test_model_covariate.model_df.copy()

    model_df.loc[:, "branch"] = ["treatment-a"] * model_df.shape[0]

    with pytest.raises(
        MissingModelParameter, match="Effect for branch control not found in model!"
    ):
        mafslm.fit_model(
            model_df,
            test_model_covariate.target,
            test_model_covariate.ref_branch,
            test_model_covariate.treatment_branches,
            test_model_covariate.covariate,
        )


def test__create_datagrid():
    # without covariate
    branches = test_model.treatment_branches + [test_model.ref_branch]
    newdata = func._create_datagrid(deepcopy(test_model.results), branches)

    expected_data = []
    for branch in branches:
        expected_data.append(
            {
                "branch": branch,
            }
        )
    expected = pl.DataFrame(expected_data)
    pl_assert_frame_equal(newdata.select("branch"), expected)

    # with covariate
    branches = test_model_covariate.treatment_branches + [
        test_model_covariate.ref_branch
    ]
    newdata = func._create_datagrid(
        test_model_covariate.results, branches, test_model_covariate.covariate
    )
    expected_data = []
    for branch in branches:
        for q in test_model_covariate.model_df[test_model_covariate.covariate].quantile(
            np.arange(0, 1, 0.0001)
        ):
            expected_data.append({"branch": branch, test_model_covariate.covariate: q})
    expected = pl.DataFrame(expected_data)

    pl_assert_frame_equal(
        newdata.select("branch", test_model_covariate.covariate).sort(
            "branch", test_model_covariate.covariate
        ),
        expected.sort("branch"),
    )

    # handle pandas types
    branches = test_model_covariate.treatment_branches + [
        test_model_covariate.ref_branch
    ]
    model_df = test_model_covariate.model_df.copy()

    model_df["search_count"] = model_df.search_count.astype(int).astype(pd.Int64Dtype())
    model_df[test_model_covariate.covariate] = (
        model_df[test_model_covariate.covariate].astype(int).astype(pd.Int64Dtype())
    )

    results = smf.ols(test_model.formula, model_df).fit()
    # should not throw, even though types are pandas types
    func._create_datagrid(results, branches, test_model_covariate.covariate)


def test_relative_inferences_with_without_datagrid():
    from marginaleffects import avg_comparisons

    branches = test_model.treatment_branches + [test_model.ref_branch]
    datagrid = func._create_datagrid(deepcopy(test_model.results), branches)

    for branch in test_model.treatment_branches:
        expected = avg_comparisons(
            test_model.results,
            variables={"branch": [test_model.ref_branch, branch]},
            comparison="lnratioavg",
            transform=np.exp,
            conf_level=1 - 0.05,
        )

        assert expected.shape == (
            1,
            7,
        ), "avg_comparisons result object not shaped as expected"

        actual = avg_comparisons(
            test_model.results,
            variables={"branch": [test_model.ref_branch, branch]},
            comparison="lnratioavg",
            transform=np.exp,
            conf_level=1 - 0.05,
            newdata=datagrid,
        )

        assert actual.shape == (
            1,
            7,
        ), "avg_comparisons result object not shaped as expected"

        assert np.isclose(expected["conf_low"][0], actual["conf_low"][0])
        assert np.isclose(expected["estimate"][0], actual["estimate"][0])
        assert np.isclose(expected["conf_high"][0], actual["conf_high"][0])


def test_relative_inferences_with_without_datagrid_covariate():
    from marginaleffects import avg_comparisons

    branches = test_model_covariate.treatment_branches + [
        test_model_covariate.ref_branch
    ]
    datagrid = func._create_datagrid(deepcopy(test_model_covariate.results), branches)

    for branch in test_model_covariate.treatment_branches:
        expected = avg_comparisons(
            test_model_covariate.results,
            variables={"branch": [test_model_covariate.ref_branch, branch]},
            comparison="lnratioavg",
            transform=np.exp,
            conf_level=1 - 0.05,
        )

        assert expected.shape == (
            1,
            7,
        ), "avg_comparisons result object not shaped as expected"

        actual = avg_comparisons(
            test_model_covariate.results,
            variables={"branch": [test_model_covariate.ref_branch, branch]},
            comparison="lnratioavg",
            transform=np.exp,
            conf_level=1 - 0.05,
            newdata=datagrid,
        )

        assert actual.shape == (
            1,
            7,
        ), "avg_comparisons result object not shaped as expected"

        assert np.isclose(expected["conf_low"][0], actual["conf_low"][0])
        assert np.isclose(expected["estimate"][0], actual["estimate"][0])
        assert np.isclose(expected["conf_high"][0], actual["conf_high"][0])


def test__validate_parameters():
    with pytest.raises(ValueError, match="Target metric fake_var not found in data"):
        func._validate_parameters(
            test_model.model_df, "fake_var", test_model.ref_branch, None, None, None
        )

    df = test_model.model_df.copy()
    df["search_count"] = 1
    with pytest.raises(
        func.UnableToAnalyze, match="Metric search_count has no variation"
    ):
        func._validate_parameters(
            df, "search_count", test_model.ref_branch, None, None, None
        )

    with pytest.raises(
        ValueError, match="No data from reference branch fake_branch found"
    ):
        func._validate_parameters(
            test_model.model_df, test_model.target, "fake_branch", None, None, None
        )

    msg = re.escape("Threshold quantile must be in (0,1]")
    with pytest.raises(ValueError, match=msg):
        func._validate_parameters(
            test_model.model_df,
            test_model.target,
            test_model.ref_branch,
            None,
            -1,
            None,
        )

    with pytest.raises(ValueError, match=msg):
        func._validate_parameters(
            test_model.model_df,
            test_model.target,
            test_model.ref_branch,
            None,
            0,
            None,
        )

    with pytest.raises(ValueError, match=msg):
        func._validate_parameters(
            test_model.model_df,
            test_model.target,
            test_model.ref_branch,
            None,
            1.00000001,
            None,
        )

    msg = re.escape("alpha must be in (0.002,1)")
    with pytest.raises(ValueError, match=msg):
        func._validate_parameters(
            test_model.model_df,
            test_model.target,
            test_model.ref_branch,
            None,
            None,
            [0.001],
        )

    with pytest.raises(ValueError, match=msg):
        func._validate_parameters(
            test_model.model_df,
            test_model.target,
            test_model.ref_branch,
            None,
            None,
            [1.001],
        )

    with pytest.raises(ValueError, match=msg):
        func._validate_parameters(
            test_model.model_df,
            test_model.target,
            test_model.ref_branch,
            None,
            None,
            [0.05, 0.1, 0.01, 0.000001],
        )

    with pytest.raises(
        ValueError,
        match=f"Covariate {test_model_covariate.target} must be different than target {test_model_covariate.target}",  # noqa: E501
    ):
        func._validate_parameters(
            test_model_covariate.model_df,
            test_model_covariate.target,
            test_model_covariate.ref_branch,
            test_model_covariate.target,
            None,
            None,
        )

    with pytest.raises(
        func.CovariateNotFound, match="Covariate fake_covariate not found in data"
    ):
        func._validate_parameters(
            test_model_covariate.model_df,
            test_model_covariate.target,
            test_model_covariate.ref_branch,
            "fake_covariate",
            None,
            None,
        )


def test__make_empty_compare_branches_output():
    df = pd.DataFrame({"branch": ["control", "treatment-a", "treatment-b"]})
    str_quantiles = ["0.5", "exp", "0.025", "0.975"]
    index = pd.MultiIndex.from_tuples(
        product(["abs_uplift", "rel_uplift"], str_quantiles)
    )
    estimates = pd.Series(index=index)

    treatment_branches = ["treatment-a", "treatment-b"]

    actual = func._make_empty_compare_branches_output(
        "control", df, [0.05], treatment_branches
    )

    for branch_df in actual["comparative"].values():
        pd.testing.assert_series_equal(estimates.sort_index(), branch_df.sort_index())



from .classes import MozOLS  # noqa: F401
from .functions import (
    compare_branches_lm,  # noqa: F401
    fit_model,  # noqa: F401
    make_formula,  # noqa: F401
    prepare_df_for_modeling,  # noqa: F401
    summarize_joint,  # noqa: F401
    summarize_one_branch,  # noqa: F401
    summarize_univariate,  # noqa: F401
)



from typing import Literal

import numpy as np
import pandas as pd
from formulaic import Formula, ModelMatrices
from formulaic.materializers import PandasMaterializer
from statsmodels.regression.linear_model import (
    OLS,
    OLSResults,
    RegressionResults,
    RegressionResultsWrapper,
)


class MozOLS(OLS):
    @classmethod
    def from_formula(
        cls, formula: str, data: pd.DataFrame, *args, **kwargs
    ) -> "MozOLS":
        """
        Create a Model from a formula and dataframe. This uses Formulaic instead
        of Patsy to create sparse design matrices which can dramatically reduce
        memory requirements (especially in experiments with many branches).

        Parameters
        ----------
        formula : str
            The formula specifying the model.
        data : pd.DataFrame
            The data for the model.
        *args
            Additional positional argument that are passed to the model.
        **kwargs
            Additional positional argument that are passed to the model. `missing`
        and `hasconst` are overridden.
        """

        mm = PandasMaterializer(data).get_model_matrix(Formula(formula))

        assert isinstance(
            mm, ModelMatrices
        ), "Expected to create multiple matrices, is the formula correct?"

        kwargs.update(
            {
                "formula": formula,
                "design_info": mm.rhs.model_spec,
                "missing": "none",
                "hasconst": False,
            }
        )

        mod = cls(mm.lhs, mm.rhs, *args, **kwargs)
        mod.formula = formula
        mod.data.frame = data
        mod.k_constant = 1
        mod.const_idx = 0

        return mod

    def fit(
        self,
        cov_type: Literal[
            "nonrobust",
            "fixed scale",
            "HC0",
            "HC1",
            "HC2",
            "HC3",
            "HAC",
            "hac-panel",
            "hac-groupsum",
            "cluster",
        ] = "nonrobust",
        cov_kwds=None,
        use_t: bool | None = None,
    ) -> RegressionResults:
        """
        Fits the model by solving the normal equations. This is more memory efficient
        than the pinv or QR methods available in statsmodels for our particular design
        matrix structure: (n,m) with n >> m. In particular, n can be >= 1e7 while m is
        1 + (n branches - 1) + 1 (if using covariate). When running in prod with the
        built in fitting algorithms, Jetstream can run out of memory.

        See section 4, specifically algorithm 4.1 from:
        https://math.uchicago.edu/~may/REU2012/REUPapers/Lee.pdf

        See also: https://xkcd.com/1838/

        Parameters
        ----------
        use_t : bool, optional
            Flag indicating to use the Student's t distribution when computing
            p-values.  Default behavior depends on cov_type. See
            `linear_model.RegressionResults.get_robustcov_results` for
            implementation details.
        cov_type : str, optional
            See `regression.linear_model.RegressionResults` for a description
            of the available covariance estimators.
        cov_kwds : list or None, optional
            See `linear_model.RegressionResults.get_robustcov_results` for a
            description required keywords for alternative covariance
            estimators.

        Returns
        -------
        RegressionResults
            The model estimation results.
        """
        y, X = self.wendog, self.wexog

        C = np.dot(X.T, X)
        # may throw np.linalg.LinAlgError if columns are collinear
        C_inv = np.linalg.inv(C)
        self.normalized_cov_params = C_inv

        L = np.linalg.cholesky(C)
        d = np.dot(X.T, y)
        sol = np.linalg.solve(L, d)
        beta = np.linalg.solve(L.T, sol)

        if self.rank is None:
            self.rank = np.linalg.matrix_rank(C)
        if self._df_model is None:
            self._df_model = float(self.rank - self.k_constant)
        if self._df_resid is None:
            self.df_resid = self.nobs - self.rank

        lfit = OLSResults(
            self,
            beta,
            normalized_cov_params=self.normalized_cov_params,
            cov_type=cov_type,
            cov_kwds=cov_kwds,
            use_t=use_t,
        )

        return RegressionResultsWrapper(lfit)
